import re
import math
from xml.etree import ElementTree

from .handler import SvgHandler, NameMode
from ... import objects
from ...nvector import NVector
from ...utils import restructure
from ...utils.transform import TransformMatrix
from ...parsers import glaxnimate_helpers
from ...utils.color import Color, ColorMode
try:
    from ...utils import font
    has_font = True
except ImportError:
    has_font = False


_supported_font_weights = {
    "Thin": 100, "Hairline": 100,
    "ExtraLight": 200, "UltraLight": 200,
    "Light": 300,
    "Regular": 400, "Normal": 400, "Plain": 400, "Standard": 400, "Roman": 400,
    "Medium": 500,
    "SemiBold": 600, "Demi": 600, "DemiBold": 600,
    "Bold": 700,
    "Extra": 800, "ExtraBold": 800, "Ultra": 800, "UltraBold": 800,
    "Black": 900, "Heavy": 900,
    "ExtraBlack": 1000, "UltraBlack": 1000, "UltraHeavy": 1000,
}


class PrecompTime:
    def __init__(self, pcl: objects.PreCompLayer):
        self.pcl = pcl

    def get_time_offset(self, time, lot):
        remap = time
        if self.pcl.time_remapping:
            remapf = self.pcl.time_remapping.get_value(time)
            remap = lot.in_point * (1-remapf) + lot.out_point * remapf

        return remap - self.pcl.start_time


class SvgBuilder(SvgHandler, restructure.AbstractBuilder):
    merge_paths = True
    namestart = (
        r":_A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF" +
        r"\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF" +
        r"\uFDF0-\uFFFD\U00010000-\U000EFFFF"
    )
    namenostart = r"-.0-9\xB7\u0300-\u036F\u203F-\u2040"
    id_re = re.compile("^[%s][%s%s]*$" % (namestart, namenostart, namestart))

    def __init__(self, time=0):
        super().__init__()
        self.svg = ElementTree.Element("svg")
        self.dom = ElementTree.ElementTree(self.svg)
        self.svg.attrib["xmlns"] = self.ns_map["svg"]
        self.ids = set()
        self.idc = 0
        self.name_mode = NameMode.Inkscape
        self.actual_time = time
        self.precomp_times = []
        self._precomps = {}
        self._assets = {}
        self._fonts = {}
        self._current_layer = []

    @property
    def time(self):
        time = self.actual_time
        if self.precomp_times:
            for pct in self.precomp_times:
                time = pct.get_time_offset(time, self._current_layer[-1])
        return time

    def gen_id(self, prefix="id"):
        while True:
            self.idc += 1
            id = "%s_%s" % (prefix, self.idc)
            if id not in self.ids:
                break
        self.ids.add(id)
        return id

    def set_clean_id(self, dom, n):
        idn = n.replace(" ", "_")
        if self.id_re.match(idn) and idn not in self.ids:
            self.ids.add(idn)
        else:
            idn = self.gen_id(dom.tag)

        dom.attrib["id"] = idn
        return idn

    def set_id(self, dom, lottieobj, inkscape_qual=None, force=False):
        n = getattr(lottieobj, "name", None)
        if n is None or self.name_mode == NameMode.NoName:
            if force:
                id = self.gen_id(dom.tag)
                dom.attrib["id"] = id
                return id
            return None

        idn = self.set_clean_id(dom, n)
        if inkscape_qual is None:
            inkscape_qual = self.qualified("inkscape", "label")
        if inkscape_qual:
            dom.attrib[inkscape_qual] = n
        return idn

    def _on_animation(self, animation: objects.Animation):
        self.svg.attrib["width"] = str(animation.width)
        self.svg.attrib["height"] = str(animation.height)
        self.svg.attrib["viewBox"] = "0 0 %s %s" % (animation.width, animation.height)
        self.svg.attrib["version"] = "1.1"
        self.set_id(self.svg, animation, self.qualified("sodipodi", "docname"))
        self.defs = ElementTree.SubElement(self.svg, "defs")
        if self.name_mode == NameMode.Inkscape:
            self.svg.attrib[self.qualified("inkscape", "export-xdpi")] = "96"
            self.svg.attrib[self.qualified("inkscape", "export-ydpi")] = "96"
            namedview = ElementTree.SubElement(self.svg, self.qualified("sodipodi", "namedview"))
            namedview.attrib[self.qualified("inkscape", "pagecheckerboard")] = "true"
            namedview.attrib["borderlayer"] = "true"
            namedview.attrib["bordercolor"] = "#666666"
            namedview.attrib["pagecolor"] = "#ffffff"
        self.svg.attrib["style"] = "fill: none; stroke: none"

        self._current_layer = [animation]
        return self.svg

    def _mask_to_def(self, mask):
        svgmask = ElementTree.SubElement(self.defs, "mask")
        mask_id = self.gen_id()
        svgmask.attrib["id"] = mask_id
        svgmask.attrib["mask-type"] = "alpha"
        path = ElementTree.SubElement(svgmask, "path")
        path.attrib["d"] = self._bezier_to_d(mask.shape.get_value(self.time))
        path.attrib["fill"] = "#fff"
        path.attrib["fill-opacity"] = str(mask.opacity.get_value(self.time) / 100)
        return mask_id

    def _matte_source_to_def(self, layer_builder):
        svgmask = ElementTree.SubElement(self.defs, "mask")
        if not layer_builder.matte_id:
            layer_builder.matte_id = self.gen_id()
        svgmask.attrib["id"] = layer_builder.matte_id
        matte_mode = layer_builder.matte_target.lottie.matte_mode

        mask_type = "alpha"
        if matte_mode == objects.MatteMode.Luma:
            mask_type = "luminance"
        svgmask.attrib["mask-type"] = mask_type
        return svgmask

    def _on_masks(self, masks):
        if len(masks) == 1:
            return self._mask_to_def(masks[0])
        mask_ids = list(map(self._mask_to_def, masks))
        mask_def = ElementTree.SubElement(self.defs, "mask")
        mask_id = self.gen_id()
        mask_def.attrib["id"] = mask_id
        g = mask_def
        for mid in mask_ids:
            g = ElementTree.SubElement(g, "g")
            g.attrib["mask"] = "url(#%s)" % mid
        full = ElementTree.SubElement(g, "rect")
        full.attrib["fill"] = "#fff"
        full.attrib["width"] = self.svg.attrib["width"]
        full.attrib["height"] = self.svg.attrib["height"]
        full.attrib["x"] = "0"
        full.attrib["y"] = "0"
        return mask_id

    def _on_layer(self, layer_builder, dom_parent):
        lot = layer_builder.lottie
        self._current_layer.append(lot)

        if not self.precomp_times and (lot.in_point > self.time or lot.out_point < self.time):
            self._current_layer.pop()
            return None

        if layer_builder.matte_target:
            dom_parent = self._matte_source_to_def(layer_builder)

        g = self.group_from_lottie(lot, dom_parent, True)

        if lot.masks:
            g.attrib["mask"] = "url(#%s)" % self._on_masks(lot.masks)
        elif layer_builder.matte_source:
            matte_id = layer_builder.matte_source.matte_id
            if not matte_id:
                matte_id = layer_builder.matte_source.matte_id = self.gen_id()
            g.attrib["mask"] = "url(#%s)" % matte_id

        if isinstance(lot, objects.PreCompLayer):
            self.precomp_times.append(PrecompTime(lot))

            for layer in self._precomps.get(lot.reference_id, []):
                self.process_layer(layer, g)

            self.precomp_times.pop()
        elif isinstance(lot, objects.NullLayer):
            g.attrib["opacity"] = "1"
        elif isinstance(lot, objects.ImageLayer):
            use = ElementTree.SubElement(g, "use")
            use.attrib[self.qualified("xlink", "href")] = "#" + self._assets[lot.image_id]
        elif isinstance(lot, objects.TextLayer):
            self._on_text_layer(g, lot)
        elif isinstance(lot, objects.SolidColorLayer):
            rect = ElementTree.SubElement(g, "rect")
            rect.attrib["width"] = str(lot.width)
            rect.attrib["height"] = str(lot.height)
            rect.attrib["fill"] = color_to_css(lot.color)

        if not lot.name:
            g.attrib[self.qualified("inkscape", "label")] = lot.__class__.__name__
        if layer_builder.shapegroup:
            g.attrib["style"] = self.group_to_style(layer_builder.shapegroup)
            self._split_stroke(layer_builder.shapegroup, g, dom_parent)
        #if lot.hidden:
            #g.attrib.setdefault("style", "")
            #g.attrib["style"] += "display: none;"

        return g

    def _on_font(self, font):
        self._fonts[font.name] = {
            "font-family": font.font_family,
            "font-weight": str(_supported_font_weights.get(font.font_style, 400)),
        }

    def _on_text_layer(self, g, lot):
        text = ElementTree.SubElement(g, "text")
        doc = lot.data.get_value(self.time)
        if doc:
            text.attrib.update(self._fonts.get(doc.font_family, {}))
            text.attrib["font-size"] = str(doc.font_size)
            if doc.line_height:
                text.attrib["line-height"] = "%s%%" % doc.line_height
            if doc.justify == objects.text.TextJustify.Left:
                text.attrib["text-anchor"] = "start"
            elif doc.justify == objects.text.TextJustify.Center:
                text.attrib["text-anchor"] = "middle"
            elif doc.justify == objects.text.TextJustify.Right:
                text.attrib["text-anchor"] = "end"

            text.attrib["fill"] = color_to_css(doc.color)
            text.text = doc.text

    def _on_layer_end(self, out_layer):
        self._current_layer.pop()

    def _on_precomp(self, id, dom_parent, layers):
        self._precomps[id] = layers

    def _on_asset(self, asset):
        if isinstance(asset, objects.assets.Image):
            img = ElementTree.SubElement(self.defs, "image")
            xmlid = self.set_clean_id(img, asset.id)
            self._assets[asset.id] = xmlid
            if asset.is_embedded:
                url = asset.image
            else:
                url = asset.path + asset.file_name
            img.attrib[self.qualified("xlink", "href")] = url
            img.attrib["width"] = str(asset.width)
            img.attrib["height"] = str(asset.height)

    def _get_value(self, prop, default=NVector(0, 0)):
        if prop:
            v = prop.get_value(self.time)
        else:
            v = default

        if v is None:
            return default
        if isinstance(v, NVector):
            return v.clone()
        return v

    def set_transform(self, dom, transform, auto_orient=False):
        if not transform:
            return

        mat = transform.to_matrix(self.time, auto_orient)
        dom.attrib["transform"] = mat.to_css_2d()

        if transform.opacity is not None:
            op = transform.opacity.get_value(self.time)
            if op != 100:
                dom.attrib["opacity"] = str(op/100)

    def _get_group_stroke(self, group):
        style = {}
        if group.stroke:
            if isinstance(group.stroke, objects.GradientStroke):
                style["stroke"] = "url(#%s)" % self.process_gradient(group.stroke)
            else:
                style["stroke"] = color_to_css(group.stroke.color.get_value(self.time))

            style["stroke-opacity"] = group.stroke.opacity.get_value(self.time) / 100
            style["stroke-width"] = group.stroke.width.get_value(self.time)
            if group.stroke.miter_limit is not None:
                style["stroke-miterlimit"] = group.stroke.miter_limit

            if group.stroke.line_cap == objects.LineCap.Round:
                style["stroke-linecap"] = "round"
            elif group.stroke.line_cap == objects.LineCap.Butt:
                style["stroke-linecap"] = "butt"
            elif group.stroke.line_cap == objects.LineCap.Square:
                style["stroke-linecap"] = "square"

            if group.stroke.line_join == objects.LineJoin.Round:
                style["stroke-linejoin"] = "round"
            elif group.stroke.line_join == objects.LineJoin.Bevel:
                style["stroke-linejoin"] = "bevel"
            elif group.stroke.line_join == objects.LineJoin.Miter:
                style["stroke-linejoin"] = "miter"

            if group.stroke.dashes:
                dasharray = []
                last = 0
                last_mode = objects.StrokeDashType.Dash
                for dash in group.stroke.dashes:
                    if last_mode == dash.type:
                        last += dash.length.get_value(self.time)
                    else:
                        if last_mode != objects.StrokeDashType.Offset:
                            dasharray.append(str(last))
                            last = 0
                        last_mode = dash.type
                style["stroke-dasharray"] = " ".join(dasharray)
        return style

    def _style_to_css(self, style):
        return ";".join(map(
            lambda x: ":".join(map(str, x)),
            style.items()
        ))

    def _split_stroke(self, group, fill_layer, out_parent):
        if not group.stroke:# or group.stroke_above:
            return

        style = self._get_group_stroke(group)
        if style.get("stroke-width", 0) <= 0 or style["stroke-opacity"] <= 0:
            return

        if group.stroke_above:
            if fill_layer.attrib.get("style", ""):
                fill_layer.attrib["style"] += ";"
            else:
                fill_layer.attrib["style"] = ""
            fill_layer.attrib["style"] += self._style_to_css(style)
            return fill_layer

        g = ElementTree.Element("g")
        self.set_clean_id(g, "stroke")
        use = ElementTree.Element("use")
        for i, e in enumerate(out_parent):
            if e is fill_layer:
                out_parent.insert(i, g)
                out_parent.remove(fill_layer)
                break
        else:
            return

        g.append(use)
        g.append(fill_layer)

        use.attrib[self.qualified("xlink", "href")] = "#" + fill_layer.attrib["id"]
        use.attrib["style"] = self._style_to_css(style)
        return g

    def group_to_style(self, group):
        style = {}
        if group.fill:
            style["fill-opacity"] = group.fill.opacity.get_value(self.time) / 100
            if isinstance(group.fill, objects.GradientFill):
                style["fill"] = "url(#%s)" % self.process_gradient(group.fill)
            else:
                style["fill"] = color_to_css(group.fill.color.get_value(self.time))

            if group.fill.fill_rule:
                style["fill-rule"] = "evenodd" if group.fill.fill_rule == objects.FillRule.EvenOdd else "nonzero"

        if group.lottie.hidden:
            style["display"] = "none"
        #if group.stroke_above:
            #style.update(self._get_group_stroke(group))

        return self._style_to_css(style)

    def process_gradient(self, gradient):
        spos = gradient.start_point.get_value(self.time)
        epos = gradient.end_point.get_value(self.time)

        if gradient.gradient_type == objects.GradientType.Linear:
            dom = ElementTree.SubElement(self.defs, "linearGradient")
            dom.attrib["x1"] = str(spos[0])
            dom.attrib["y1"] = str(spos[1])
            dom.attrib["x2"] = str(epos[0])
            dom.attrib["y2"] = str(epos[1])
        elif gradient.gradient_type == objects.GradientType.Radial:
            dom = ElementTree.SubElement(self.defs, "radialGradient")
            dom.attrib["cx"] = str(spos[0])
            dom.attrib["cy"] = str(spos[1])
            dom.attrib["r"] = str((epos-spos).length)
            a = gradient.highlight_angle.get_value(self.time) * math.pi / 180
            l = gradient.highlight_length.get_value(self.time)
            dom.attrib["fx"] = str(spos[0] + math.cos(a) * l)
            dom.attrib["fy"] = str(spos[1] + math.sin(a) * l)

        id = self.set_id(dom, gradient, force=True)
        dom.attrib["gradientUnits"] = "userSpaceOnUse"

        for off, color in gradient.colors.stops_at(self.time):
            stop = ElementTree.SubElement(dom, "stop")
            stop.attrib["offset"] = "%s%%" % (off * 100)
            stop.attrib["stop-color"] = color_to_css(color[:3])
            if len(color) > 3:
                stop.attrib["stop-opacity"] = str(color[3])

        return id

    def group_from_lottie(self, lottie, dom_parent, layer):
        g = ElementTree.SubElement(dom_parent, "g")
        if layer and self.name_mode == NameMode.Inkscape:
            g.attrib[self.qualified("inkscape", "groupmode")] = "layer"
        self.set_id(g, lottie, force=True)
        self.set_transform(g, lottie.transform, getattr(lottie, "auto_orient", False))
        return g

    def _on_shapegroup(self, group, dom_parent):
        if group.empty():
            return

        if len(group.children) == 1 and isinstance(group.children[0], restructure.RestructuredPathMerger):
            path = self.build_path(group.paths.paths, dom_parent)
            self.set_id(path, group.paths.paths[0], force=True)
            path.attrib["style"] = self.group_to_style(group)
            self.set_transform(path, group.lottie.transform)
            return self._split_stroke(group, path, dom_parent)

        g = self.group_from_lottie(group.lottie, dom_parent, group.layer)
        g.attrib["style"] = self.group_to_style(group)
        self.shapegroup_process_children(group, g)
        return self._split_stroke(group, g, dom_parent)

    def _on_merged_path(self, shape, shapegroup, out_parent):
        path = self.build_path(shape.paths, out_parent)
        self.set_id(path, shape.paths[0])
        path.attrib["style"] = self.group_to_style(shapegroup)
        #self._split_stroke(shapegroup, path, out_parent)
        return path

    def _on_shape(self, shape, shapegroup, out_parent):
        if isinstance(shape, objects.Rect):
            svgshape = self.build_rect(shape, out_parent)
        elif isinstance(shape, objects.Ellipse):
            svgshape = self.build_ellipse(shape, out_parent)
        elif isinstance(shape, objects.Star):
            svgshape = self.build_path([shape.to_bezier()], out_parent)
        elif isinstance(shape, objects.Path):
            svgshape = self.build_path([shape], out_parent)
        elif has_font and isinstance(shape, font.FontShape):
            svgshape = self.build_text(shape, out_parent)
        else:
            return
        self.set_id(svgshape, shape, force=True)
        if "style" not in svgshape.attrib:
            svgshape.attrib["style"] = ""
        svgshape.attrib["style"] += self.group_to_style(shapegroup)
        #self._split_stroke(shapegroup, svgshape, out_parent)

        if shape.hidden:
            svgshape.attrib["style"] += "display: none;"
        return svgshape

    def build_rect(self, shape, parent):
        rect = ElementTree.SubElement(parent, "rect")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        rect.attrib["width"] = str(size[0])
        rect.attrib["height"] = str(size[1])
        rect.attrib["x"] = str(pos[0] - size[0] / 2)
        rect.attrib["y"] = str(pos[1] - size[1] / 2)
        rect.attrib["rx"] = str(shape.rounded.get_value(self.time))
        return rect

    def build_ellipse(self, shape, parent):
        ellipse = ElementTree.SubElement(parent, "ellipse")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        ellipse.attrib["rx"] = str(size[0] / 2)
        ellipse.attrib["ry"] = str(size[1] / 2)
        ellipse.attrib["cx"] = str(pos[0])
        ellipse.attrib["cy"] = str(pos[1])
        return ellipse

    def build_path(self, shapes, parent):
        path = ElementTree.SubElement(parent, "path")
        d = ""
        for shape in shapes:
            bez = shape.shape.get_value(self.time)
            if isinstance(bez, list):
                bez = bez[0]
            if not bez.vertices:
                continue
            if d:
                d += "\n"
            d += self._bezier_to_d(bez)

        path.attrib["d"] = d
        return path

    def _bezier_tangent(self, tangent):
        _tangent_threshold = 0.5
        if tangent.length < _tangent_threshold:
            return NVector(0, 0)
        return tangent

    def _bezier_to_d(self, bez):
        d = "M %s,%s " % tuple(bez.vertices[0].components[:2])
        for i in range(1, len(bez.vertices)):
            qfrom = bez.vertices[i-1]
            h1 = self._bezier_tangent(bez.out_tangents[i-1]) + qfrom
            qto = bez.vertices[i]
            h2 = self._bezier_tangent(bez.in_tangents[i]) + qto

            d += "C %s,%s %s,%s %s,%s " % (
                h1[0], h1[1],
                h2[0], h2[1],
                qto[0], qto[1],
            )
        if bez.closed:
            qfrom = bez.vertices[-1]
            h1 = self._bezier_tangent(bez.out_tangents[-1]) + qfrom
            qto = bez.vertices[0]
            h2 = self._bezier_tangent(bez.in_tangents[0]) + qto
            d += "C %s,%s %s,%s %s,%s Z" % (
                h1[0], h1[1],
                h2[0], h2[1],
                qto[0], qto[1],
            )

        return d

    def _on_shape_modifier(self, shape, shapegroup, out_parent):
        if isinstance(shape.lottie, objects.Repeater):
            svgshape = self.build_repeater(shape.lottie, shape.child, shapegroup, out_parent)
        elif isinstance(shape.lottie, objects.RoundedCorners):
            svgshape = self.build_rouded_corners(shape.lottie, shape.child, shapegroup, out_parent)
        elif isinstance(shape.lottie, objects.Trim):
            svgshape = self.build_trim_path(shape.lottie, shape.child, shapegroup, out_parent)
        else:
            return self.shapegroup_process_child(shape.child, shapegroup, out_parent)
        return svgshape

    def build_repeater(self, shape, child, shapegroup, out_parent):
        original = self.shapegroup_process_child(child, shapegroup, out_parent)
        if not original:
            return

        ncopies = int(round(shape.copies.get_value(self.time)))
        if ncopies == 1:
            return

        out_parent.remove(original)

        g = ElementTree.SubElement(out_parent, "g")
        self.set_clean_id(g, "repeater")

        for copy in range(ncopies-1):
            use = ElementTree.SubElement(g, "use")
            use.attrib[self.qualified("xlink", "href")] = "#" + original.attrib["id"]

        orig_wrapper = ElementTree.SubElement(g, "g")
        orig_wrapper.append(original)

        transform = objects.Transform()
        so = shape.transform.start_opacity.get_value(self.time)
        eo = shape.transform.end_opacity.get_value(self.time)
        position = shape.transform.position.get_value(self.time)
        rotation = shape.transform.rotation.get_value(self.time)
        anchor_point = shape.transform.anchor_point.get_value(self.time)
        for i in range(ncopies-1, -1, -1):
            of = i / (ncopies-1)
            transform.opacity.value = so * of + eo * (1 - of)
            self.set_transform(g[i], transform)
            transform.position.value += position
            transform.rotation.value += rotation
            transform.anchor_point.value += anchor_point

        return g

    def build_rouded_corners(self, shape, child, shapegroup, out_parent):
        round_amount = shape.radius.get_value(self.time)
        return self._modifier_process(child, shapegroup, out_parent, self._build_rouded_corners_shape, round_amount)

    def _build_rouded_corners_shape(self, shape, round_amount):
        if not isinstance(shape, objects.Shape):
            return [shape]
        path = shape.to_bezier()
        bezier = path.shape.get_value(self.time).rounded(round_amount)
        path.shape.clear_animation(bezier)
        return [path]

    def build_trim_path(self, shape, child, shapegroup, out_parent):
        start = max(0, min(1, shape.start.get_value(self.time) / 100))
        end = max(0, min(1, shape.end.get_value(self.time) / 100))
        offset = shape.offset.get_value(self.time) / 360 % 1

        multidata = {}
        length = 0

        if shape.multiple == objects.TrimMultipleShapes.Individually:
            for visishape in reversed(list(self._modifier_foreach_shape(child))):
                bez = visishape.to_bezier().shape.get_value(self.time)
                local_length = bez.rough_length()
                multidata[visishape] = (bez, length, local_length)
                length += local_length

        return self._modifier_process(
            child, shapegroup, out_parent, self._build_trim_path_shape,
            start+offset, end+offset, multidata, length
        )

    def _modifier_foreach_shape(self, shape):
        if isinstance(shape, restructure.RestructuredShapeGroup):
            for child in shape.children:
                for chsh in self._modifier_foreach_shape(child):
                    yield chsh
        elif isinstance(shape, restructure.RestructuredPathMerger):
            for p in shape.paths:
                yield p
        elif isinstance(shape, objects.Shape):
            yield shape

    def _modifier_process(self, child, shapegroup, out_parent, callback, *args):
        children = self._modifier_process_child(child, shapegroup, out_parent, callback, *args)
        return [self.shapegroup_process_child(ch, shapegroup, out_parent) for ch in children]

    def _trim_offlocal(self, t, local_start, local_length, total_length):
        gt = (t * total_length - local_start) / local_length
        return max(0, min(1, gt))

    def _build_trim_path_shape(self, shape, start, end, multidata, total_length):
        if not isinstance(shape, objects.Shape):
            return [shape]

        if multidata:
            bezier, local_start, local_length = multidata[shape]
            if end > 1:
                lstart = self._trim_offlocal(start, local_start, local_length, total_length)
                lend = self._trim_offlocal(end-1, local_start, local_length, total_length)
                out = []
                if lstart < 1:
                    out.append(objects.Path(bezier.segment(lstart, 1)))
                if lend > 0:
                    out.append(objects.Path(bezier.segment(0, lend)))
                return out

            lstart = self._trim_offlocal(start, local_start, local_length, total_length)
            lend = self._trim_offlocal(end, local_start, local_length, total_length)
            if lend <= 0 or lstart >= 1:
                return []
            if lstart <= 0 and lend >= 1:
                return [objects.Path(bezier)]
            seg = bezier.segment(lstart, lend)
            return [objects.Path(seg)]

        path = shape.to_bezier()
        bezier = path.shape.get_value(self.time)
        if end > 1:
            bez1 = bezier.segment(start, 1)
            bez2 = bezier.segment(0, end-1)
            return [objects.Path(bez1), objects.Path(bez2)]
        else:
            seg = bezier.segment(start, end)
            return [objects.Path(seg)]

    def _modifier_process_children(self, shapegroup, out_parent, callback, *args):
        children = []
        for shape in shapegroup.children:
            children.extend(self._modifier_process_child(shape, shapegroup, out_parent, callback, *args))
        shapegroup.children = children

    def _modifier_process_child(self, shape, shapegroup, out_parent, callback, *args):
        if isinstance(shape, restructure.RestructuredShapeGroup):
            self._modifier_process_children(shape, out_parent, callback, *args)
            return [shape]
        elif isinstance(shape, restructure.RestructuredPathMerger):
            paths = []
            for p in shape.paths:
                paths.extend(callback(p, *args))
            shape.paths = paths
            if paths:
                return [shape]
            return []
        else:
            return callback(shape, *args)

    def _custom_object_supported(self, shape):
        if has_font and isinstance(shape, font.FontShape):
            return True
        return False

    def build_text(self, shape, parent):
        text = ElementTree.SubElement(parent, "text")
        if "family" in shape.query:
            text.attrib["font-family"] = shape.query["family"]
        if "weight" in shape.query:
            text.attrib["font-weight"] = str(shape.query.weight_to_css())
        slant = int(shape.query.get("slant", 0))
        if slant > 0 and slant < 110:
            text.attrib["font-style"] = "italic"
        elif slant >= 110:
            text.attrib["font-style"] = "oblique"

        text.attrib["font-size"] = str(shape.size)

        text.attrib["white-space"] = "pre"

        pos = shape.style.position
        text.attrib["x"] = str(pos.x)
        text.attrib["y"] = str(pos.y)
        text.text = shape.text

        return text


def color_to_css(color):
    #if len(color) == 4:
        #return ("rgba(%s, %s, %s" % tuple(map(lambda c: int(round(c*255)), color[:3]))) + ", %s)" % color[3]
    if isinstance(color, Color) and color.mode != ColorMode.RGB:
        color = color.converted(ColorMode.RGB)
    return "rgb(%s, %s, %s)" % tuple(map(lambda c: int(round(c*255)), color[:3]))


def to_svg(animation, time, animated=False):
    if animated:
        data = glaxnimate_helpers.convert(animation, "svg")
        return ElementTree.ElementTree(ElementTree.fromstring(data.decode("utf8")))

    if glaxnimate_helpers.has_glaxnimate:
        data = glaxnimate_helpers.serialize(animation, "svg")
        return ElementTree.ElementTree(ElementTree.fromstring(data.decode("utf8")))

    builder = SvgBuilder(time)
    builder.process(animation)
    return builder.dom



from . import base, core, sif, svg, pretty_print

from .base import exporters
from .core import export_lottie, export_tgs, export_embedded_html
from .pretty_print import prettyprint, prettyprint_summary
from .sif import export_sif
from .svg import export_svg

__all__ = [
    "base", "core", "sif", "svg", "pretty_print",
    "exporters", "export_lottie", "export_tgs", "export_embedded_html",
    "prettyprint", "prettyprint_summary", "export_sif", "export_svg",
]

try:
    from . import cairo, gif
    __all__ += ["cairo", "gif"]
except ImportError:
    pass



import math
from .base import LottieObject, LottieProp, LottieEnum
from .properties import MultiDimensional, Value, NVector, ShapeProperty, PositionValue


## @ingroup Lottie
class Transform(LottieObject):
    """!
    Layer transform
    """
    _props = [
        LottieProp("anchor_point", "a", PositionValue, False),
        LottieProp("position", "p", PositionValue, False),
        LottieProp("scale", "s", MultiDimensional, False),
        LottieProp("rotation", "r", Value, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("skew", "sk", Value, False),
        LottieProp("skew_axis", "sa", Value, False),
        LottieProp("orientation", "or", MultiDimensional, False),
    ]

    def __init__(self):
        ## Transform Anchor Point
        self.anchor_point = PositionValue(NVector(0, 0))
        ## Transform Position
        self.position = PositionValue(NVector(0, 0))
        ## Transform Scale
        self.scale = MultiDimensional(NVector(100, 100))
        ## Transform Rotation
        self.rotation = Value(0)
        ## Transform Opacity
        self.opacity = Value(100)

        """
        # Transform Position X
        #self.position_x = Value()
        ## Transform Position Y
        #self.position_y = Value()
        ## Transform Position Z
        #self.position_z = Value()
        """

        ## Transform Skew
        self.skew = Value(0)
        ## Transform Skew Axis.
        ## An angle, if 0 skews on the X axis, if 90 skews on the Y axis
        self.skew_axis = Value(0)

        self.orientation = None

    def to_matrix(self, time, auto_orient=False):
        from ..utils.transform import TransformMatrix
        mat = TransformMatrix()

        anchor = self.anchor_point.get_value(time) if self.anchor_point else NVector(0, 0)
        mat.translate(-anchor.x, -anchor.y)

        scale = self.scale.get_value(time) if self.scale else NVector(100, 100)
        mat.scale(scale.x / 100, scale.y / 100)

        skew = (self.skew.get_value(time) * math.pi / 180) if self.skew else 0
        if skew != 0:
            axis = (self.skew_axis.get_value(time) * math.pi / 180) if self.skew_axis else 0
            mat.skew_from_axis(-skew, axis)

        rot = (self.rotation.get_value(time) * math.pi / 180) if self.rotation else 0
        if rot:
            mat.rotate(-rot)

        if auto_orient:
            if self.position and self.position.animated:
                ao_angle = self.position.get_tangent_angle(time)
                mat.rotate(-ao_angle)

        pos = self.position.get_value(time) if self.position else NVector(0, 0)
        mat.translate(pos.x, pos.y)

        return mat


## @ingroup Lottie
class MaskMode(LottieEnum):
    """!
    How masks interact with each other
    @see https://helpx.adobe.com/after-effects/using/alpha-channels-masks-mattes.html
    """
    No = "n"
    Add = "a"
    Subtract = "s"
    Intersect = "i"
    ## @note Not in lottie web
    Lighten = "l"
    ## @note Not in lottie web
    Darken = "d"
    ## @note Not in lottie web
    Difference = "f"


## @ingroup Lottie
class VisualObject(LottieObject):
    _props = [
        LottieProp("name", "nm", str, False),
        LottieProp("match_name", "mn", str, False),
    ]

    def __init__(self):
        super().__init__()
        # Object name
        self.name = None
        # Used for expression
        self.match_name = None


## @ingroup Lottie
## @todo Implement SVG/SIF I/O
class Mask(VisualObject):
    _props = [
        LottieProp("inverted", "inv", bool, False),
        LottieProp("shape", "pt", ShapeProperty, False),
        LottieProp("opacity", "o", Value, False),
        LottieProp("mode", "mode", MaskMode, False),
        LottieProp("dilate", "x", Value, False),
    ]

    def __init__(self, bezier=None):
        super().__init__()
        ## Inverted Mask flag
        self.inverted = False
        ## Mask vertices
        self.shape = ShapeProperty(bezier)
        ## Mask opacity.
        self.opacity = Value(100)
        ## Mask mode. Not all mask types are supported.
        self.mode = MaskMode.Intersect
        self.dilate = Value(0)

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
class BlendMode(LottieEnum):
    Normal = 0
    Multiply = 1
    Screen = 2
    Overlay = 3
    Darken = 4
    Lighten = 5
    ColorDodge = 6
    ColorBurn = 7
    HardLight = 8
    SoftLight = 9
    Difference = 10
    Exclusion = 11
    Hue = 12
    Saturation = 13
    Color = 14
    Luminosity = 15


#ingroup Lottie
class Marker(LottieObject):
    """!
    Defines named portions of the composition.
    """
    _props = [
        LottieProp("comment", "cm", str, False),
        LottieProp("time", "tm", float, False),
        LottieProp("duration", "dr", float, False),
    ]

    def __init__(self):
        super().__init__()

        self.comment = None
        self.time = None
        self.duration = None



#!/usr/bin/env python3

import json
import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.exporters import prettyprint, prettyprint_summary
from lottie.objects import Animation
from lottie.parsers.tgs import parse_tgs
from lottie import __version__

parser = argparse.ArgumentParser(
    description="Pretty prints a tgs / lottie file with more readable annotations (useful to debug / diff lottie files)"
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "infile",
    help="Input file"
)
parser.add_argument(
    "--summary",
    "-s",
    action="store_true",
    help="Just show a short summary"
)

if __name__ == "__main__":
    ns = parser.parse_args()

    an = parse_tgs(ns.infile)

    if ns.summary:
        prettyprint_summary(an)
    else:
        prettyprint(an)




import io
from PIL import Image
from PIL import features

from .cairo import PngRenderer
from .base import exporter, io_progress
from ..parsers.baseporter import ExtraOption


def _png_gif_prepare(image):
    if image.mode not in ["RGBA", "RGBa"]:
        image = image.convert("RGBA")
    alpha = image.getchannel("A")
    image = image.convert("RGB").convert('P', palette=Image.ADAPTIVE, colors=255)
    mask = Image.eval(alpha, lambda a: 255 if a <= 128 else 0)
    image.paste(255, mask=mask)
    return image


def _log_frame(fmt, frame_no=None, end=None):
    if frame_no is None:
        io_progress().report_message("%s frame rendering completed" % (fmt))
    else:
        io_progress().report_progress("%s rendering frame" % fmt, frame_no, end)


@exporter("GIF", ["gif"], [
    ExtraOption("skip_frames", type=int, default=1, help="Only renderer 1 out of these many frames"),
])
def export_gif(animation, fp, dpi=96, skip_frames=1):
    """
    Gif export

    Note that it's a bit slow.
    """
    start = int(animation.in_point)
    end = int(animation.out_point)
    frames = []
    with PngRenderer(animation, dpi) as renderer:
        for i in range(start, end+1, skip_frames):
            _log_frame("GIF", i, end)
            file = io.BytesIO()
            renderer.serialize(i, file)
            file.seek(0)
            frames.append(_png_gif_prepare(Image.open(file)))
    _log_frame("GIF")

    io_progress().report_message("GIF Writing to file...")
    duration = int(round(1000 / animation.frame_rate * skip_frames / 10)) * 10
    frames[0].save(
        fp,
        format='GIF',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        transparency=255,
        disposal=2,
    )


@exporter("WebP", ["webp"], [
    ExtraOption("lossless", action="store_true", help="If present, use lossless compression"),
    ExtraOption("quality", type=int, default=80,
                help="Compression effort between 0 and 100\n" +
                     "for lossy 0 gives the smallest size\n" +
                     "for lossless 0 gives the largest file"),
    ExtraOption("method", type=int, default=0, help="Quality/speed trade-off (0=fast, 6=slower-better)"),
    ExtraOption("skip_frames", type=int, default=1, help="Only renderer 1 out of these many frames"),
])
def export_webp(animation, fp, dpi=96, lossless=False, quality=80, method=0, skip_frames=1):
    """
    Export WebP

    See https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#webp
    """
    if not features.check("webp_anim"):
        raise Exception("WebP animations not supported in this system")

    start = int(animation.in_point)
    end = int(animation.out_point)
    frames = []
    with PngRenderer(animation, dpi) as renderer:
        for i in range(start, end+1, skip_frames):
            _log_frame("WebP", i, end)
            file = io.BytesIO()
            renderer.serialize(i, file)
            file.seek(0)
            frames.append(Image.open(file))

    _log_frame("WebP")

    io_progress().report_message("WebP Writing to file...")
    duration = int(round(1000 / animation.frame_rate * skip_frames))
    frames[0].save(
        fp,
        format='WebP',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        background=(0, 0, 0, 0),
        lossless=lossless,
        quality=quality,
        method=method
    )


@exporter("TIFF", ["tiff"])
def export_tiff(animation, fp, dpi=96):
    """
    Export TIFF
    """
    start = int(animation.in_point)
    end = int(animation.out_point)
    frames = []
    with PngRenderer(animation, dpi) as renderer:
        for i in range(start, end+1):
            _log_frame("TIFF", i, end)
            file = io.BytesIO()
            renderer.serialize(i, file)
            file.seek(0)
            frames.append(Image.open(file))
    _log_frame("TIFF")

    io_progress().report_message("TIFF Writing to file...")
    duration = int(round(1000 / animation.frame_rate))
    frames[0].save(
        fp,
        format='TIFF',
        append_images=frames[1:],
        save_all=True,
        duration=duration,
        loop=0,
        dpi=(dpi, dpi),
    )



import os
import sys
import subprocess
import fontTools.pens.basePen
import fontTools.ttLib
import fontTools.t1Lib
from fontTools.pens.boundsPen import ControlBoundsPen
import enum
import math
from xml.etree import ElementTree
from ..nvector import NVector
from ..objects.bezier import Bezier, BezierPoint
from ..objects.shapes import Path, Group, Fill, Stroke
from ..objects.text import TextJustify
from ..objects.base import LottieProp, CustomObject
from ..objects.layers import ShapeLayer


class BezierPen(fontTools.pens.basePen.BasePen):
    def __init__(self, glyphSet, offset=NVector(0, 0)):
        super().__init__(glyphSet)
        self.beziers = []
        self.current = Bezier()
        self.offset = offset

    def _point(self, pt):
        return self.offset + NVector(pt[0], -pt[1])

    def _moveTo(self, pt):
        self._endPath()

    def _endPath(self):
        if len(self.current.points):
            self.beziers.append(self.current)
        self.current = Bezier()

    def _closePath(self):
        self.current.close()
        self._endPath()

    def _lineTo(self, pt):
        if len(self.current.points) == 0:
            self.current.points.append(self._point(self._getCurrentPoint()))

        self.current.points.append(self._point(pt))

    def _curveToOne(self, pt1, pt2, pt3):
        if len(self.current.points) == 0:
            cp = self._point(self._getCurrentPoint())
            self.current.points.append(
                BezierPoint(
                    cp,
                    None,
                    self._point(pt1) - cp
                )

            )
        else:
            self.current.points[-1].out_tangent = self._point(pt1) - self.current.points[-1].vertex

        dest = self._point(pt3)
        self.current.points.append(
            BezierPoint(
                dest,
                self._point(pt2) - dest,
                None,
            )
        )


class SystemFont:
    def __init__(self, family):
        self.family = family
        self.files = {}
        self.styles = set()
        self._renderers = {}

    def add_file(self, styles, file):
        self.styles |= set(styles)
        key = self._key(styles)
        self.files.setdefault(key, file)

    def filename(self, styles):
        return self.files[self._key(styles)]

    def _key(self, styles):
        if isinstance(styles, str):
            return (styles,)
        return tuple(sorted(styles))

    def __getitem__(self, styles):
        key = self._key(styles)
        if key in self._renderers:
            return self._renderers[key]
        fr = RawFontRenderer(self.files[key])
        self._renderers[key] = fr
        return fr

    def __repr__(self):
        return "<SystemFont %s>" % self.family


class FontQuery:
    """!
    @see https://www.freedesktop.org/software/fontconfig/fontconfig-user.html#AEN21
         https://manpages.ubuntu.com/manpages/cosmic/man1/fc-pattern.1.html
    """
    def __init__(self, str=""):
        self._query = {}
        if isinstance(str, FontQuery):
            self._query = str._query.copy()
        elif str:
            chunks = str.split(":")
            family = chunks.pop(0)
            self._query = dict(
                chunk.split("=")
                for chunk in chunks
                if chunk
            )
            self.family(family)

    def family(self, name):
        self._query["family"] = name
        return self

    def weight(self, weight):
        self._query["weight"] = weight
        return self

    def css_weight(self, weight):
        """!
            Weight from CSS weight value.

            Weight is different between CSS and fontconfig
            This creates some interpolations to ensure known values are translated properly
            @see https://www.freedesktop.org/software/fontconfig/fontconfig-user.html#AEN178
                 https://developer.mozilla.org/en-US/docs/Web/CSS/font-weight#Common_weight_name_mapping
        """
        if weight < 200:
            v = max(0, weight - 100) / 100 * 40
        elif weight < 500:
            v = -weight**3 / 200000 + weight**2 * 11/2000 - weight * 17/10 + 200
        elif weight < 700:
            v = -weight**2 * 3/1000 + weight * 41/10 - 1200
        else:
            v = (weight - 700) / 200 * 10 + 200
        return self.weight(int(round(v)))

    def style(self, *styles):
        self._query["style"] = " ".join(styles)
        return self

    def charset(self, *hex_ranges):
        self._query["charset"] = " ".join(hex_ranges)
        return self

    def char(self, char):
        return self.charset("%x" % ord(char))

    def custom(self, property, value):
        self._query[property] = value
        return self

    def clone(self):
        return FontQuery(self)

    def __getitem__(self, key):
        return self._query.get(key, "")

    def __contains__(self, item):
        return item in self._query

    def get(self, key, default=None):
        return self._query.get(key, default)

    def __str__(self):
        return self._query.get("family", "") + ":" + ":".join(
            "%s=%s" % (p, v)
            for p, v in self._query.items()
            if p != "family"
        )

    def __repr__(self):
        return "<FontQuery %r>" % str(self)

    def weight_to_css(self):
        x = int(self["weight"])
        if x < 40:
            v = x / 40 * 100 + 100
        elif x < 100:
            v = x**3/300 - x**2 * 11/15 + x*167/3 - 3200/3
        elif x < 200:
            v = (2050 - 10 * math.sqrt(5) * math.sqrt(1205 - 6 * x)) / 3
        else:
            v = (x - 200) * 200 / 10 + 700
        return int(round(v))


class _SystemFontList:
    def __init__(self):
        self.fonts = None

    def _lazy_load(self):
        if self.fonts is None:
            self.load()

    def load(self):
        self.fonts = {}
        self.load_fc_list()

    def cmd(self, *a):
        p = subprocess.Popen(a, stdout=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode("utf-8").strip()
        return out, p.returncode

    def load_fc_list(self):
        out, returncode = self.cmd("fc-list", r'--format=%{file}\t%{family[0]}\t%{style[0]}\n')
        if returncode == 0:
            for line in out.splitlines():
                file, family, styles = line.split("\t")
                self._get(family).add_file(styles.split(" "), file)

    def best(self, query):
        """!
        Returns the renderer best matching the name
        """
        out, returncode = self.cmd("fc-match", r"--format=%{family}\t%{style}", str(query))
        if returncode == 0:
            return self._font_from_match(out)

    def _font_from_match(self, out):
        fam, style = out.split("\t")
        fam = fam.split(",")[0]
        style = style.split(",")[0].split()
        return self[fam][style]

    def all(self, query):
        """!
        Yields all the renderers matching a query
        """
        out, returncode = self.cmd("fc-match", "-s", r"--format=%{family}\t%{style}\n", str(query))
        if returncode == 0:
            for line in out.splitlines():
                try:
                    yield self._font_from_match(line)
                except (fontTools.ttLib.TTLibError, fontTools.t1Lib.T1Error):
                    pass

    def default(self):
        """!
        Returns the default fornt renderer
        """
        return self.best()

    def _get(self, family):
        self._lazy_load()
        if family in self.fonts:
            return self.fonts[family]
        font = SystemFont(family)
        self.fonts[family] = font
        return font

    def __getitem__(self, key):
        self._lazy_load()
        return self.fonts[key]

    def __iter__(self):
        self._lazy_load()
        return iter(self.fonts.values())

    def keys(self):
        self._lazy_load()
        return self.fonts.keys()

    def __contains__(self, item):
        self._lazy_load()
        return item in self.fonts


## Dictionary of system fonts
fonts = _SystemFontList()


def collect_kerning_pairs(font):
    if "GPOS" not in font:
        return {}

    gpos_table = font["GPOS"].table

    unique_kern_lookups = set()
    for item in gpos_table.FeatureList.FeatureRecord:
        if item.FeatureTag == "kern":
            feature = item.Feature
            unique_kern_lookups |= set(feature.LookupListIndex)

    kerning_pairs = {}
    for kern_lookup_index in sorted(unique_kern_lookups):
        lookup = gpos_table.LookupList.Lookup[kern_lookup_index]
        if lookup.LookupType in {2, 9}:
            for pairPos in lookup.SubTable:
                if pairPos.LookupType == 9:  # extension table
                    if pairPos.ExtensionLookupType == 8:  # contextual
                        continue
                    elif pairPos.ExtensionLookupType == 2:
                        pairPos = pairPos.ExtSubTable

                if pairPos.Format != 1:
                    continue

                firstGlyphsList = pairPos.Coverage.glyphs
                for ps_index, _ in enumerate(pairPos.PairSet):
                    for pairValueRecordItem in pairPos.PairSet[ps_index].PairValueRecord:
                        secondGlyph = pairValueRecordItem.SecondGlyph
                        valueFormat = pairPos.ValueFormat1

                        if valueFormat == 5:  # RTL kerning
                            kernValue = "<%d 0 %d 0>" % (
                                pairValueRecordItem.Value1.XPlacement,
                                pairValueRecordItem.Value1.XAdvance)
                        elif valueFormat == 0:  # RTL pair with value <0 0 0 0>
                            kernValue = "<0 0 0 0>"
                        elif valueFormat == 4:  # LTR kerning
                            kernValue = pairValueRecordItem.Value1.XAdvance
                        else:
                            print(
                                "\tValueFormat1 = %d" % valueFormat,
                                file=sys.stdout)
                            continue  # skip the rest

                        kerning_pairs[(firstGlyphsList[ps_index], secondGlyph)] = kernValue
    return kerning_pairs


class GlyphMetrics:
    def __init__(self, glyph, lsb, aw, xmin, xmax):
        self.glyph = glyph
        self.lsb = lsb
        self.advance = aw
        self.xmin = xmin
        self.xmax = xmax
        self.width = xmax - xmin
        self.advance = xmax

    def draw(self, pen):
        return self.glyph.draw(pen)


class Font:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        if isinstance(self.wrapped, fontTools.ttLib.TTFont):
            self.cmap = self.wrapped.getBestCmap() or {}
        else:
            self.cmap = {}

        self.glyphset = self.wrapped.getGlyphSet()

    @classmethod
    def open(cls, filename):
        try:
            f = fontTools.ttLib.TTFont(filename)
        except fontTools.ttLib.TTLibError:
            f = fontTools.t1Lib.T1Font(filename)
            f.parse()

        return cls(f)

    def getGlyphSet(self):
        return self.wrapped.getGlyphSet()

    def getBestCmap(self):
        return {}

    def glyph_name(self, codepoint):
        if isinstance(codepoint, str):
            if len(codepoint) != 1:
                return ""
            codepoint = ord(codepoint)

        if codepoint in self.cmap:
            return self.cmap[codepoint]

        return self.calculated_glyph_name(codepoint)

    @staticmethod
    def calculated_glyph_name(codepoint):
        from fontTools import agl  # Adobe Glyph List
        if codepoint in agl.UV2AGL:
            return agl.UV2AGL[codepoint]
        elif codepoint <= 0xFFFF:
            return "uni%04X" % codepoint
        else:
            return "u%X" % codepoint

    def scale(self):
        if isinstance(self.wrapped, fontTools.ttLib.TTFont):
            return 1 / self.wrapped["head"].unitsPerEm
        elif isinstance(self.wrapped, fontTools.t1Lib.T1Font):
            return self.wrapped["FontMatrix"][0]

    def yMax(self):
        if isinstance(self.wrapped, fontTools.ttLib.TTFont):
            return self.wrapped["head"].yMax
        elif isinstance(self.wrapped, fontTools.t1Lib.T1Font):
            return self.wrapped["FontBBox"][3]

    def glyph(self, glyph_name):
        if isinstance(self.wrapped, fontTools.ttLib.TTFont):
            glyph = self.glyphset[glyph_name]

            xmin = getattr(glyph._glyph, "xMin", glyph.lsb)
            xmax = getattr(glyph._glyph, "xMax", glyph.width)
            return GlyphMetrics(glyph, glyph.lsb, glyph.width, xmin, xmax)
        elif isinstance(self.wrapped, fontTools.t1Lib.T1Font):
            glyph = self.glyphset[glyph_name]
            bounds_pen = ControlBoundsPen(self.glyphset)
            bounds = bounds_pen.bounds
            glyph.draw(bounds_pen)
            if not hasattr(glyph, "width"):
                advance = bounds[2]
            else:
                advance = glyph.width
            return GlyphMetrics(glyph, bounds[0], advance, bounds[0], bounds[2])

    def __contains__(self, key):
        if isinstance(self.wrapped, fontTools.t1Lib.T1Font):
            return key in self.wrapped.font
        return key in self.wrapped

    def __getitem__(self, key):
        return self.wrapped[key]


class FontRenderer:
    tab_width = 4

    @property
    def font(self):
        raise NotImplementedError

    def get_query(self):
        raise NotImplementedError

    def kerning(self, c1, c2):
        return 0

    def text_to_chars(self, text):
        return text

    def _on_missing(self, char, size, pos, group):
        """!
        - Character as string
        - Font size
        - [in, out] Character position
        - Group shape
        """

    def glyph_name(self, ch):
        return self.font.glyph_name(ch)

    def scale(self, size):
        return size * self.font.scale()

    def line_height(self, size):
        return self.font.yMax() * self.scale(size)

    def ex(self, size):
        return self.font.glyph("x").advance * self.scale(size)

    def glyph_beziers(self, glyph, offset=NVector(0, 0)):
        pen = BezierPen(self.font.glyphset, offset)
        glyph.draw(pen)
        return pen.beziers

    def glyph_shapes(self, glyph, offset=NVector(0, 0)):
        beziers = self.glyph_beziers(glyph, offset)
        return [
            Path(bez)
            for bez in beziers
        ]

    def _on_character(self, ch, size, pos, scale, line, use_kerning, chars, i):
        chname = self.glyph_name(ch)

        if chname in self.font.glyphset:
            glyphdata = self.font.glyph(chname)
            #pos.x += glyphdata.lsb * scale
            glyph_shapes = self.glyph_shapes(glyphdata, pos / scale)

            if glyph_shapes:
                if len(glyph_shapes) > 1:
                    glyph_shape_group = line.add_shape(Group())
                    glyph_shape = glyph_shape_group
                else:
                    glyph_shape_group = line
                    glyph_shape = glyph_shapes[0]

                for sh in glyph_shapes:
                    sh.shape.value.scale(scale)
                    glyph_shape_group.add_shape(sh)

                glyph_shape.name = ch

            kerning = 0
            if use_kerning and i < len(chars) - 1:
                nextcname = chars[i+1]
                kerning = self.kerning(chname, nextcname)

            pos.x += (glyphdata.advance + kerning) * scale
            return True
        return False

    def render(self, text, size, pos=None, use_kerning=True, start_x=None):
        """!
        Renders some text

        @param text         String to render
        @param size         Font size (in pizels)
        @param[in,out] pos  Text position
        @param use_kerning  Whether to honour kerning info from the font file
        @param start_x      x-position of the start of a line

        @returns a Group shape, augmented with some extra attributes:
        - line_height   Line height
        - next_x        X position of the next character
        """
        scale = self.scale(size)
        line_height = self.line_height(size)
        group = Group()
        group.name = text
        if pos is None:
            pos = NVector(0, 0)
        start_x = pos.x if start_x is None else start_x
        line = Group()
        group.add_shape(line)
        #group.transform.scale.value = NVector(100, 100) * scale

        chars = self.text_to_chars(text)
        for i, ch in enumerate(chars):
            if ch == "\n":
                line.next_x = pos.x
                pos.x = start_x
                pos.y += line_height
                line = Group()
                group.add_shape(line)
                continue
            elif ch == "\t":
                chname = self.glyph_name(ch)
                if chname in self.font.glyphset:
                    width = self.font.glyph(chname).advance
                else:
                    width = self.ex(size)
                pos.x += width * scale * self.tab_width
                continue

            self._on_character(ch, size, pos, scale, line, use_kerning, chars, i)

        group.line_height = line_height
        group.next_x = line.next_x = pos.x
        return group


class RawFontRenderer(FontRenderer):
    def __init__(self, filename):
        self.filename = filename
        self._font = Font.open(filename)
        self._kerning = None

    @property
    def font(self):
        return self._font

    def kerning(self, c1, c2):
        if self._kerning is None:
            self._kerning = collect_kerning_pairs(self.font)
        return self._kerning.get((c1, c2), 0)

    def __repr__(self):
        return "<FontRenderer %r>" % self.filename

    def get_query(self):
        return self.filename


class FallbackFontRenderer(FontRenderer):
    def __init__(self, query, max_attempts=10):
        self.query = FontQuery(query)
        self._best = None
        self._bq = None
        self._fallback = {}
        self.max_attempts = max_attempts

    @property
    def font(self):
        return self.best.font

    def get_query(self):
        return self.query

    def ex(self, size):
        best = self.best
        if "x" not in self.font.glyphset:
            best = fonts.best(self.query.clone().char("x"))
        return best.ex(size)

    @property
    def best(self):
        cq = str(self.query)
        if self._best is None or self._bq != cq:
            self._best = fonts.best(self.query)
            self._bq = cq
        return self._best

    def fallback_renderer(self, char):
        if char in self._fallback:
            return self._fallback[char]

        if len(char) != 1:
            return None

        codepoint = ord(char)
        name = Font.calculated_glyph_name(codepoint)
        for i, font in enumerate(fonts.all(self.query.clone().char(char))):
            # For some reason fontconfig sometimes returns a font that doesn't
            # actually contain the glyph
            if name in font.font.glyphset or codepoint in font.font.cmap:
                self._fallback[char] = font
                return font

            if i > self.max_attempts:
                self._fallback[char] = None
                return None

    def _on_character(self, char, size, pos, scale, group, use_kerning, chars, i):
        if self.best._on_character(char, size, pos, scale, group, use_kerning, chars, i):
            return True

        font = self.fallback_renderer(char)
        if not font:
            return False

        child = font.render(char, size, pos)
        if len(child.shapes) == 2:
            group.add_shape(child.shapes[0])
        else:
            group.add_shape(child)

    def __repr__(self):
        return "<FallbackFontRenderer %s>" % self.query


class EmojiRenderer(FontRenderer):
    _split = None

    def __init__(self, wrapped, emoji_dir):
        if not os.path.isdir(emoji_dir):
            raise Exception("Not a valid directory: %s" % emoji_dir)
        self.wrapped = wrapped
        self.emoji_dir = emoji_dir
        self._svgs = {}

    @property
    def font(self):
        return self.wrapped.font

    def emoji_basename(self, char):
        return "-".join("%x" % ord(cp) for cp in char)

    def emoji_filename(self, char):
        return self._get_svg_filename(char)[1]

    def _get_svg_filename(self, char):
        basename = self.emoji_basename(char)
        suffix = ".svg"

        filename = os.path.join(self.emoji_dir, basename + suffix)
        if os.path.isfile(filename):
            return basename, filename

        filename = os.path.join(self.emoji_dir, basename.upper() + suffix)
        if os.path.isfile(filename):
            return basename, filename

        if char and char[-1] == '\ufe0f':
            return self._get_svg_filename(char[:-1])

        return None, None

    def _get_svg(self, char):
        from ..parsers.svg import parse_svg_file

        if char in self._svgs:
            return self._svgs[char]

        basename, filename = self._get_svg_filename(char)
        if filename is None:
            self._svgs[char] = None
            return None

        svga = parse_svg_file(filename)
        svgshape = Group()
        svgshape.name = basename
        for layer in svga.layers:
            if isinstance(layer, ShapeLayer):
                for shape in layer.shapes:
                    svgshape.add_shape(shape)

        self._svgs[char] = svgshape
        svgshape._bbox = svgshape.bounding_box()
        return svgshape

    def _on_character(self, char, size, pos, scale, group, use_kerning, chars, i):
        svgshape = self._get_svg(char)
        if svgshape:
            target_height = self.line_height(size)
            scale = target_height / svgshape._bbox.height
            shape_group = Group()
            shape_group = svgshape.clone()
            shape_group.transform.scale.value *= scale
            offset = NVector(
                -svgshape._bbox.x1 + svgshape._bbox.width * 0.075,
                -svgshape._bbox.y2 + svgshape._bbox.height * 0.1
            )
            shape_group.transform.position.value = pos + offset * scale
            group.add_shape(shape_group)
            pos.x += svgshape._bbox.width * scale
            return True
        return self.wrapped._on_character(char, size, pos, scale, group, use_kerning, chars, i)

    def get_query(self):
        return self.wrapped.get_query()

    @staticmethod
    def _get_splitter():
        if EmojiRenderer._split is None:
            try:
                import grapheme
                EmojiRenderer._split = grapheme.graphemes
            except ImportError:
                sys.stderr.write("Install `grapheme` for better Emoji support\n")
                EmojiRenderer._split = lambda x: x
        return EmojiRenderer._split

    @staticmethod
    def emoji_split(string):
        return EmojiRenderer._get_splitter()(string)

    def text_to_chars(self, string):
        return list(self.emoji_split(string))


class FontStyle:
    def __init__(self, query, size, justify=TextJustify.Left, position=None, use_kerning=True, emoji_svg=None):
        self.emoji_svg = emoji_svg
        self._set_query(query)
        self.size = size
        self.justify = justify
        self.position = position.clone() if position else NVector(0, 0)
        self.use_kerning = use_kerning

    def _set_query(self, query):
        if isinstance(query, str) and os.path.isfile(query):
            self._renderer = RawFontRenderer(query)
        else:
            self._renderer = FallbackFontRenderer(query)

        if self.emoji_svg:
            self._renderer = EmojiRenderer(self._renderer, self.emoji_svg)

    @property
    def query(self):
        return self._renderer.get_query()

    @query.setter
    def query(self, value):
        if str(value) != str(self.query):
            self._set_query(value)

    @property
    def renderer(self):
        return self._renderer

    def render(self, text, pos=NVector(0, 0)):
        group = self._renderer.render(text, self.size, self.position+pos, self.use_kerning)
        for subg in group.shapes[:-1]:
            width = subg.next_x - self.position.x - pos.x
            if self.justify == TextJustify.Center:
                subg.transform.position.value.x -= width / 2
            elif self.justify == TextJustify.Right:
                subg.transform.position.value.x -= width
        return group

    def clone(self):
        return FontStyle(str(self.query), self.size, self.justify, NVector(*self.position), self.use_kerning)

    @property
    def ex(self):
        return self._renderer.ex(self.size)

    @property
    def line_height(self):
        return self._renderer.line_height(self.size)


def _propfac(a):
    return property(lambda s: s._get(a), lambda s, v: s._set(a, v))


class FontShape(CustomObject):
    _props = [
        LottieProp("query_string", "_query", str),
        LottieProp("size", "_size", float),
        LottieProp("justify", "_justify", TextJustify),
        LottieProp("text", "_text", str),
        LottieProp("position", "_position", NVector),
    ]
    wrapped_lottie = Group

    def __init__(self, text="", query="", size=64, justify=TextJustify.Left):
        CustomObject.__init__(self)
        if isinstance(query, FontStyle):
            self.style = query
        else:
            self.style = FontStyle(query, size, justify)
        self.text = text
        self.hidden = None

    def _get(self, a):
        return getattr(self.style, a)

    def _set(self, a, v):
        return setattr(self.style, a, v)

    query = _propfac("query")
    size = _propfac("size")
    justify = _propfac("justify")
    position = _propfac("position")

    @property
    def query_string(self):
        return str(self.query)

    @query_string.setter
    def query_string(self, v):
        self.query = v

    def _build_wrapped(self):
        g = self.style.render(self.text)
        self.line_height = g.line_height
        return g

    def bounding_box(self, time=0):
        return self.wrapped.bounding_box(time)



import random
import math
from ..nvector import NVector
from ..objects.shapes import Path
from .. import objects
from ..objects import easing
from ..objects import properties


def shake(position_prop, x_radius, y_radius, start_time, end_time, n_frames, interp=easing.Linear()):
    if not isinstance(position_prop, list):
        position_prop = [position_prop]

    n_frames = int(round(n_frames))
    frame_time = (end_time - start_time) / n_frames
    startpoints = list(map(
        lambda pp: pp.get_value(start_time),
        position_prop
    ))

    for i in range(n_frames):
        x = (random.random() * 2 - 1) * x_radius
        y = (random.random() * 2 - 1) * y_radius
        for pp, start in zip(position_prop, startpoints):
            px = start[0] + x
            py = start[1] + y
            pp.add_keyframe(start_time + i * frame_time, NVector(px, py), interp)

    for pp, start in zip(position_prop, startpoints):
        pp.add_keyframe(end_time, start, interp)


def rot_shake(rotation_prop, angles, start_time, end_time, n_frames):
    frame_time = (end_time - start_time) / n_frames
    start = rotation_prop.get_value(start_time)

    for i in range(0, n_frames):
        a = angles[i % len(angles)] * math.sin(i/n_frames * math.pi)
        rotation_prop.add_keyframe(start_time + i * frame_time, start + a)
    rotation_prop.add_keyframe(end_time, start)


def spring_pull(position_prop, point, start_time, end_time, falloff=15, oscillations=7):
    start = position_prop.get_value(start_time)
    d = start-point

    delta = (end_time - start_time) / oscillations

    for i in range(oscillations):
        time_x = i / oscillations
        factor = math.cos(time_x * math.pi * oscillations) * (1-time_x**(1/falloff))
        p = point + d * factor
        position_prop.add_keyframe(start_time + delta * i, p)

    position_prop.add_keyframe(end_time, point)


def follow_path(position_prop, bezier, start_time, end_time, n_keyframes,
                reverse=False, offset=NVector(0, 0), start_t=0, rotation_prop=None, rotation_offset=0):
    delta = (end_time - start_time) / (n_keyframes-1)
    fact = start_t
    factd = 1 / (n_keyframes-1)

    if rotation_prop:
        start_rot = rotation_prop.get_value(start_time) if rotation_offset is None else rotation_offset

    for i in range(n_keyframes):
        time = start_time + i * delta

        if fact > 1 + factd/2:
            fact -= 1
            if time != start_time:
                easing.Jump()(position_prop.keyframes[-1])
                if rotation_prop:
                    easing.Jump()(rotation_prop.keyframes[-1])

        f = 1 - fact if reverse else fact
        position_prop.add_keyframe(time, bezier.point_at(f)+offset)

        if rotation_prop:
            rotation_prop.add_keyframe(time, bezier.tangent_angle_at(f) / math.pi * 180 + start_rot)

        fact += factd


def generate_path_appear(bezier, appear_start, appear_end, n_keyframes, reverse=False):
    obj = Path()
    beziers = []
    maxp = 0

    time_delta = (appear_end - appear_start) / n_keyframes
    for i in range(n_keyframes+1):
        time = appear_start + i * time_delta
        t2 = (time - appear_start) / (appear_end - appear_start)

        if reverse:
            t2 = 1 - t2
            segment = bezier.segment(t2, 1)
            segment.reverse()
        else:
            segment = bezier.segment(0, t2)

        beziers.append(segment)
        if len(segment.vertices) > maxp:
            maxp = len(segment.vertices)

        obj.shape.add_keyframe(time, segment)

    for segment in beziers:
        deltap = maxp - len(segment.vertices)
        if deltap > 0:
            segment.vertices += [segment.vertices[-1]] * deltap
            segment.in_tangents += [NVector(0, 0)] * deltap
            segment.out_tangents += [NVector(0, 0)] * deltap

    return obj


def generate_path_disappear(bezier, disappear_start, disappear_end, n_keyframes, reverse=False):
    obj = Path()
    beziers = []
    maxp = 0

    time_delta = (disappear_end - disappear_start) / n_keyframes
    for i in range(n_keyframes+1):
        time = disappear_start + i * time_delta
        t1 = (time - disappear_start) / (disappear_end - disappear_start)
        if reverse:
            t1 = 1 - t1
            segment = bezier.segment(0, t1)
        else:
            segment = bezier.segment(1, t1)
            segment.reverse()

        beziers.append(segment)
        if len(segment.vertices) > maxp:
            maxp = len(segment.vertices)

        obj.shape.add_keyframe(time, segment)

    for segment in beziers:
        deltap = maxp - len(segment.vertices)
        if deltap > 0:
            segment.vertices += [segment.vertices[-1]] * deltap
            segment.in_tangents += [NVector(0, 0)] * deltap
            segment.out_tangents += [NVector(0, 0)] * deltap

    return obj


def generate_path_segment(bezier, appear_start, appear_end, disappear_start, disappear_end, n_keyframes, reverse=False):
    obj = Path()
    beziers = []
    maxp = 0

    # HACK: For some reason reversed works better
    if not reverse:
        bezier.reverse()

    time_delta = (appear_end - appear_start) / n_keyframes
    for i in range(n_keyframes+1):
        time = appear_start + i * time_delta
        t1 = (time - disappear_start) / (disappear_end - disappear_start)
        t2 = (time - appear_start) / (appear_end - appear_start)

        t1 = max(0, min(1, t1))
        t2 = max(0, min(1, t2))

        #if reverse:
        if True:
            t1 = 1 - t1
            t2 = 1 - t2
            segment = bezier.segment(t2, t1)
            segment.reverse()
        #else:
            #segment = bezier.segment(t1, t2)
            #segment.reverse()

        beziers.append(segment)
        if len(segment.vertices) > maxp:
            maxp = len(segment.vertices)

        obj.shape.add_keyframe(time, segment)

    for segment in beziers:
        deltap = maxp - len(segment.vertices)
        if deltap > 0:
            segment.split_self_chunks(deltap+1)

    # HACK: Restore
    if not reverse:
        bezier.reverse()
    return obj


class PointDisplacer:
    def __init__(self, time_start, time_end, n_frames):
        """!
        @param time_start   When the animation shall start
        @param time_end     When the animation shall end
        @param n_frames     Number of frames in the animation
        """
        ## When the animation shall start
        self.time_start = time_start
        ## When the animation shall end
        self.time_end = time_end
        ## Number of frames in the animation
        self.n_frames = n_frames
        ## Length of a frame
        self.time_delta = (time_end - time_start) / n_frames

    def animate_point(self, prop):
        startpos = prop.get_value(self.time_start)
        for f in range(self.n_frames+1):
            p = self._on_displace(startpos, f)
            prop.add_keyframe(self.frame_time(f), startpos+p)

    def _on_displace(self, startpos, f):
        raise NotImplementedError()

    def animate_bezier(self, prop):
        initial = prop.get_value(self.time_start)

        for f in range(self.n_frames+1):
            bezier = objects.Bezier()
            bezier.closed = initial.closed

            for pi in range(len(initial.vertices)):
                startpos = initial.vertices[pi]
                dp = self._on_displace(startpos, f)
                t1sp = initial.in_tangents[pi] + startpos
                t1fin = initial.in_tangents[pi] + self._on_displace(t1sp, f) - dp
                t2sp = initial.out_tangents[pi] + startpos
                t2fin = initial.out_tangents[pi] + self._on_displace(t2sp, f) - dp

                bezier.add_point(dp + startpos, t1fin, t2fin)

            prop.add_keyframe(self.frame_time(f), bezier)

    def frame_time(self, f):
        return f * self.time_delta + self.time_start

    def _init_lerp(self, val_from, val_to, easing):
        self._kf = properties.OffsetKeyframe(0, NVector(val_from), NVector(val_to), easing)

    def _lerp_get(self, offset):
        return self._kf.interpolated_value(offset / self.n_frames)[0]


class SineDisplacer(PointDisplacer):
    def __init__(
        self,
        wavelength,
        amplitude,
        time_start,
        time_end,
        n_frames,
        speed=1,
        axis=90,
    ):
        """!
        Displaces points as if they were following a sine wave

        @param wavelength  Distance between consecutive peaks
        @param amplitude   Distance from a peak to the original position
        @param time_start  When the animation shall start
        @param time_end    When the animation shall end
        @param n_frames    Number of keyframes to add
        @param speed       Number of peaks a point will go through in the given time
                           If negative, it will go the other way
        @param axis        Wave peak direction
        """
        super().__init__(time_start, time_end, n_frames)

        self.wavelength = wavelength
        self.amplitude = amplitude
        self.speed_f = math.pi * 2 * speed
        self.axis = axis / 180 * math.pi

    def _on_displace(self, startpos, f):
        off = -math.sin(startpos[0]/self.wavelength*math.pi*2-f*self.speed_f/self.n_frames) * self.amplitude
        return NVector(off * math.cos(self.axis), off * math.sin(self.axis))


class MultiSineDisplacer(PointDisplacer):
    def __init__(
        self,
        waves,
        time_start,
        time_end,
        n_frames,
        speed=1,
        axis=90,
        amplitude_scale=1,
    ):
        """!
        Displaces points as if they were following a sine wave

        @param waves       List of tuples (wavelength, amplitude)
        @param time_start  When the animation shall start
        @param time_end    When the animation shall end
        @param n_frames    Number of keyframes to add
        @param speed       Number of peaks a point will go through in the given time
                           If negative, it will go the other way
        @param axis        Wave peak direction
        @param amplitude_scale  Multiplies the resulting amplitude by this factor
        """
        super().__init__(time_start, time_end, n_frames)

        self.waves = waves
        self.speed_f = math.pi * 2 * speed
        self.axis = axis / 180 * math.pi
        self.amplitude_scale = amplitude_scale

    def _on_displace(self, startpos, f):
        off = 0
        for wavelength, amplitude in self.waves:
            off -= math.sin(startpos[0]/wavelength*math.pi*2-f*self.speed_f/self.n_frames) * amplitude

        off *= self.amplitude_scale
        return NVector(off * math.cos(self.axis), off * math.sin(self.axis))


class DepthRotationAxis:
    def __init__(self, x, y, keep):
        self.x = x / x.length
        self.y = y / y.length
        self.keep = keep / keep.length # should be the cross product

    def rot_center(self, center, point):
        return (
            self.x * self.x.dot(center) +
            self.y * self.y.dot(center) +
            self.keep * self.keep.dot(point)
        )

    def extract_component(self, vector, axis):
        return sum(vector.element_scaled(axis).components)

    @classmethod
    def from_points(cls, keep_point, center=NVector(0, 0, 0)):
        keep = keep_point - center
        keep /= keep.length
        # Hughes-Moller to find x and y
        if abs(keep.x) > abs(keep.z):
            y = NVector(-keep.y, keep.x, 0)
        else:
            y = NVector(0, -keep.z, keep.y)
        y /= y.length
        x = y.cross(keep)
        return cls(x, y, keep)


class DepthRotation:
    axis_x = DepthRotationAxis(NVector(0, 0, 1), NVector(0, 1, 0), NVector(1, 0, 0))
    axis_y = DepthRotationAxis(NVector(1, 0, 0), NVector(0, 0, 1), NVector(0, 1, 0))
    axis_z = DepthRotationAxis(NVector(1, 0, 0), NVector(0, 1, 0), NVector(0, 0, 1))

    def __init__(self, center):
        self.center = center

    def rotate3d_y(self, point, angle):
        return self.rotate3d(point, angle, self.axis_y)
        # Hard-coded version:
        #c = NVector(self.center.x, point.y, self.center.z)
        #rad = angle * math.pi / 180
        #delta = point - c
        #pol_l = delta.length
        #pol_a = math.atan2(delta.z, delta.x)
        #dest_a = pol_a + rad
        #return NVector(
        #    c.x + pol_l * math.cos(dest_a),
        #    point.y,
        #    c.z + pol_l * math.sin(dest_a)
        #)

    def rotate3d_x(self, point, angle):
        return self.rotate3d(point, angle, self.axis_x)
        # Hard-coded version:
        #c = NVector(point.x, self.center.y, self.center.z)
        #rad = angle * math.pi / 180
        #delta = point - c
        #pol_l = delta.length
        #pol_a = math.atan2(delta.y, delta.z)
        #dest_a = pol_a + rad
        #return NVector(
        #    point.x,
        #    c.y + pol_l * math.sin(dest_a),
        #    c.z + pol_l * math.cos(dest_a),
        #)

    def rotate3d_z(self, point, angle):
        return self.rotate3d(point, angle, self.axis_z)

    def rotate3d(self, point, angle, axis):
        c = axis.rot_center(self.center, point)
        rad = angle * math.pi / 180
        delta = point - c
        pol_l = delta.length
        pol_a = math.atan2(
            axis.extract_component(delta, axis.y),
            axis.extract_component(delta, axis.x)
        )
        dest_a = pol_a + rad
        return c + axis.x * pol_l * math.cos(dest_a) + axis.y * pol_l * math.sin(dest_a)


class DepthRotationDisplacer(PointDisplacer):
    axis_x = DepthRotation.axis_x
    axis_y = DepthRotation.axis_y
    axis_z = DepthRotation.axis_z

    def __init__(self, center, time_start, time_end, n_frames, axis,
                 depth=0, angle=360, anglestart=0, ease=easing.Linear()):
        super().__init__(time_start, time_end, n_frames)
        self.rotation = DepthRotation(center)
        if isinstance(axis, NVector):
            axis = DepthRotationAxis.from_points(axis)
        self.axis = axis
        self.depth = depth
        self._angle = angle
        self.anglestart = anglestart
        self.ease = ease
        self._init_lerp(0, angle, ease)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self._init_lerp(0, value, self.ease)

    def _on_displace(self, startpos, f):
        angle = self.anglestart + self._lerp_get(f)
        if len(startpos) < 3:
            startpos = NVector(*(startpos.components + [self.depth]))
        return self.rotation.rotate3d(startpos, angle, self.axis) - startpos


class EnvelopeDeformation(PointDisplacer):
    def __init__(self, topleft, bottomright):
        self.topleft = topleft
        self.size = bottomright - topleft
        self.keyframes = []

    @property
    def time_start(self):
        return self.keyframes[0][0]

    def add_reset_keyframe(self, time):
        self.add_keyframe(
            time,
            self.topleft.clone(),
            NVector(self.topleft.x + self.size.x, self.topleft.y),
            NVector(self.topleft.x + self.size.x, self.topleft.y + self.size.y),
            NVector(self.topleft.x, self.topleft.y + self.size.y),
        )

    def add_keyframe(self, time, tl, tr, br, bl):
        self.keyframes.append([
            time,
            tl.clone(),
            tr.clone(),
            br.clone(),
            bl.clone()
        ])

    def _on_displace(self, startpos, f):
        _, tl, tr, br, bl = self.keyframes[f]
        relp = startpos - self.topleft
        relp.x /= self.size.x
        relp.y /= self.size.y

        x1 = tl.lerp(tr, relp.x)
        x2 = bl.lerp(br, relp.x)

        #return x1.lerp(x2, relp.y)
        return x1.lerp(x2, relp.y) - startpos

    @property
    def n_frames(self):
        return len(self.keyframes)-1

    def frame_time(self, f):
        return self.keyframes[f][0]


class DisplacerDampener(PointDisplacer):
    """!
    Given a displacer and a function that returns a factor for a point,
    multiplies the effect of the displacer by the factor
    """
    def __init__(self, displacer, dampener):
        self.displacer = displacer
        self.dampener = dampener

    @property
    def time_start(self):
        return self.displacer.time_start

    def _on_displace(self, startpos, f):
        disp = self.displacer._on_displace(startpos, f)
        damp = self.dampener(startpos)
        return disp * damp

    @property
    def n_frames(self):
        return self.displacer.n_frames

    def frame_time(self, f):
        return self.displacer.frame_time(f)


class FollowDisplacer(PointDisplacer):
    def __init__(
        self,
        origin,
        range,
        offset_func,
        time_start, time_end, n_frames,
        falloff_exp=1,
    ):
        """!
        @brief Uses a custom offset function, and applies a falloff to the displacement

        @param origin       Origin point for the falloff
        @param range        Radius after which the points will not move
        @param offset_func  Function returning an offset given a ratio of the time
        @param time_start   When the animation shall start
        @param time_end     When the animation shall end
        @param n_frames     Number of frames in the animation
        @param falloff_exp  Exponent for the falloff
        """
        super().__init__(time_start, time_end, n_frames)
        self.origin = origin
        self.range = range
        self.offset_func = offset_func
        self.falloff_exp = falloff_exp

    def _on_displace(self, startpos, f):
        influence = 1 - min(1, (startpos - self.origin).length / self.range) ** self.falloff_exp
        return self.offset_func(f / self.n_frames) * influence



import threading
from functools import reduce
from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5.QtCore import Qt

from ..exporters.base import exporters
from ..importers.base import importers
from ..parsers.baseporter import IoProgressReporter, ExtraOption


class GuiProgressReporter(IoProgressReporter):
    def __init__(self):
        self.dialogs = {}
        self.threads = {}
        self.lock = threading.Lock()
        self.id = 0

    @classmethod
    def set_global(cls):
        IoProgressReporter.instance = cls()

    def gen_id(self, parent):
        self.id += 1
        with self.lock:
            dialog = QtWidgets.QProgressDialog(parent)
            dialog.setCancelButton(None)
            dialog.setWindowModality(Qt.ApplicationModal)
            self.dialogs[self.id] = dialog
        return self.id

    def get_thread_dialog(self):
        with self.lock:
            id = threading.current_thread().ident
            return self.dialogs[self.threads[id]]

    def setup(self, title, id):
        with self.lock:
            tid = threading.current_thread().ident
            self.threads[tid] = id
            dialog = self.dialogs[id]

        dialog.setWindowTitle(title)
        dialog.hide()

    def report_progress(self, title, value, total):
        dialog = self.get_thread_dialog()
        dialog.show()
        dialog.setValue(value)
        dialog.setMaximum(total)
        dialog.setLabelText(title)

    def report_message(self, message):
        dialog = self.get_thread_dialog()
        dialog.show()
        dialog.setRange(0, 0)
        dialog.setLabelText(message)

    def cleanup(self):
        with self.lock:
            id = threading.current_thread().ident
            dialog = self.dialogs.pop(self.threads.pop(id))
            dialog.hide()
            dialog.deleteLater()


class GuiBasePorter:
    def __init__(self, porter):
        self.porter = porter
        self.file_filter = "%s (%s)" % (porter.name, " ".join("*.%s" % e for e in porter.extensions))
        self._dialog = None
        self.widgets = {}

    def _build_dialog(self):
        self._dialog = QtWidgets.QDialog()
        self._dialog.setWindowTitle("Export to %s" % self.porter.name)
        layout = QtWidgets.QFormLayout()
        self._dialog.setLayout(layout)

        for go in self.porter.generic_options:
            self._generic_option(layout, go)

        for option in self.porter.extra_options:
            self._add_option(layout, option)

        button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        button_box.accepted.connect(self._dialog.accept)
        button_box.rejected.connect(self._dialog.reject)
        layout.addRow(button_box)

    def _add_option(self, layout, option: ExtraOption):
        if option.dest in self.widgets:
            return
        label = QtWidgets.QLabel(option.name.replace("_", " ").title())

        default = option.kwargs.get("default", None)
        option_type = option.kwargs.get("type", str)

        if option.kwargs.get("action") == "store_true":
            widget = QtWidgets.QCheckBox()
            widget.setChecked(False)
            getter = widget.isChecked
        elif option.kwargs.get("action") == "store_false":
            widget = QtWidgets.QCheckBox()
            labtext = label.text()
            if labtext.startswith("No "):
                label.setText(labtext[3:])
            widget.setChecked(True)
            getter = widget.isChecked
        elif "choices" in option.kwargs:
            widget = QtWidgets.QComboBox()
            for choice in option.kwargs["choices"]:
                widget.addItem(str(choice))
            if default:
                widget.setCurrentText(str(default))
            getter = lambda: option_type(widget.currentText())
        elif option_type is int:
            widget = QtWidgets.QSpinBox()
            widget.setMinimum(-1000)
            widget.setMaximum(1000)
            if default is not None:
                widget.setValue(int(default))
            getter = widget.value
        else:
            widget = QtWidgets.QLineEdit()
            if default is not None:
                widget.setText(str(default))
            getter = widget.text

        help = option.kwargs.get("help", "")
        widget.setWhatsThis(help)
        widget.setToolTip(help)

        widget.setObjectName(option.dest)
        self.widgets[option.dest] = getter
        layout.addRow(label, widget)

    @property
    def needs_dialog(self):
        return self.porter.generic_options or self.porter.extra_options

    @property
    def dialog(self):
        if not self._dialog:
            self._build_dialog()
        return self._dialog

    def get_options(self):
        return {
            name: getter()
            for name, getter in self.widgets.items()
        }

    def prompt_options(self, parent):
        if self.needs_dialog:
            self.dialog.setParent(parent)
            self.dialog.setWindowFlags(Qt.Dialog)
            if self.dialog.exec_() != QtWidgets.QDialog.DialogCode.Accepted:
                return None

        return self.get_options()

    def _generic_option(self, layout, go):
        pass


class GuiExporter(GuiBasePorter):
    def _generic_option(self, layout, go):
        if go == "pretty":
            self._add_option(layout, ExtraOption("pretty", action="store_true", help="Pretty print"))
        if go == "frame":
            self._add_option(layout, ExtraOption("frame", type=int, help="Frame to extract"))

    @property
    def exporter(self):
        return self.porter


class GuiImporter(GuiBasePorter):
    @property
    def importer(self):
        return self.porter


class ExportThread(QtCore.QThread):
    def __init__(self, parent, exporter, animation, file_name, options):
        super().__init__()
        self.id = IoProgressReporter.instance.gen_id(parent)
        self.animation = animation
        self.exporter = exporter
        self.file_name = file_name
        self.options = options

    def run(self):
        IoProgressReporter.instance.setup("Export to %s" % self.exporter.name, self.id)
        try:
            self.exporter.process(self.animation, self.file_name, **self.options)
        except Exception:
            IoProgressReporter.instance.report_message("Error on export")
        IoProgressReporter.instance.cleanup()


gui_exporters = list(map(GuiExporter, exporters))
gui_importers = list(map(GuiImporter, importers))


def get_open_filename(parent, title, dirname):
    extensions = reduce(lambda a, b: a | b, (set(gi.importer.extensions) for gi in gui_importers))
    all_files = "All Supported Files (%s)" % " ".join("*.%s" % ex for ex in extensions)
    filters = all_files + ";;" + ";;".join(ge.file_filter for ge in gui_importers)
    file_name, filter = QtWidgets.QFileDialog.getOpenFileName(
        parent, title, dirname, filters
    )
    if file_name:
        if filter == all_files:
            importer = gui_importer_from_filename(file_name)
            if importer:
                return file_name, importer
        else:
            for importer in gui_importers:
                if importer.file_filter == filter:
                    return file_name, importer

    return None, None


def get_save_filename(parent, title, dirname):
    extensions = reduce(lambda a, b: a | b, (set(gi.exporter.extensions) for gi in gui_exporters))
    all_files = "All Supported Files (%s)" % " ".join("*.%s" % ex for ex in extensions)
    filters = all_files + ";;" + ";;".join(ge.file_filter for ge in gui_exporters)
    file_name, filter = QtWidgets.QFileDialog.getSaveFileName(
        parent, title, dirname, filters
    )

    if file_name:
        if filter == all_files:
            exporter = gui_exporter_from_filename(file_name)
            if exporter:
                return file_name, exporter
        else:
            for exporter in gui_exporters:
                if exporter.file_filter == filter:
                    return file_name, exporter

    return None, None


def start_export(parent, exporter, animation, file_name, options):
    thread = ExportThread(parent, exporter, animation, file_name, options)
    thread.start()
    return thread


def gui_importer_from_filename(filename):
    importer = importers.get_from_filename(filename)
    for gip in gui_importers:
        if gip.importer is importer:
            return gip
    return None


def gui_exporter_from_filename(filename):
    exporter = exporters.get_from_filename(filename)
    for gip in gui_exporters:
        if gip.exporter is exporter:
            return gip
    return None



import os
import sys
import argparse
import inspect
from ..exporters import exporters
from .stripper import float_strip


def _get_caller():
    return inspect.getmodule(inspect.currentframe().f_back.f_back)


def _get_parser(caller, basename, path, formats, verbosity):
    if basename is None:
        basename = os.path.splitext(os.path.basename(caller.__file__))[0]

    parser = argparse.ArgumentParser(
        conflict_handler='resolve'
    )

    parser.add_argument(
        "--name",
        "-n",
        default=basename,
        help="Output basename",
    )
    parser.add_argument(
        "--path",
        default=path,
        help="Output path",
    )
    parser.add_argument(
        "--formats", "-f",
        nargs="+",
        choices=list(sum((e.extensions for e in exporters), [])),
        default=formats,
        help="Formats to render",
        metavar="format"
    )
    parser.add_argument(
        "--verbosity",
        type=int,
        default=int(verbosity)
    )
    from .. import __version__
    parser.add_argument(
        "--version", "-v",
        action="version",
        version="%(prog)s - python-lottie script " + __version__
    )
    exporters.set_options(parser)

    return parser


def get_parser(basename=None, path="/tmp", formats=["html"], verbosity=1):
    caller = _get_caller()
    return _get_parser(caller, basename, path, formats, verbosity)


def run(animation, ns):
    for fmt in ns.formats:
        if ns.path == "" and ns.name == "-":
            outfile = sys.stdout
        else:
            absname = os.path.abspath(os.path.join(ns.path, ns.name + "." + fmt))
            if ns.verbosity:
                sys.stderr.write("file://%s\n" % absname)
            outfile = absname
        exporter = exporters.get_from_extension(fmt)
        exporter.process(animation, outfile, **exporter.argparse_options(ns))


def script_main(animation, basename=None, path="/tmp", formats=["html"], verbosity=1, strip=float_strip):
    """
    Sets up a script to output an animation into various formats
    """
    caller = _get_caller()
    if caller and caller.__name__ == "__main__":
        parser = _get_parser(caller, basename, path, formats, verbosity)
        strip(animation)
        run(animation, parser.parse_args())



import re
import math
import colorsys
from xml.etree import ElementTree
from ... import objects
from ...nvector import NVector
from .svgdata import color_table, css_atrrs
from .handler import SvgHandler, NameMode
from ...utils.ellipse import Ellipse
from ...utils.transform import TransformMatrix
from ...utils.color import Color

try:
    from ...utils import font
    has_font = True
except ImportError:
    has_font = False

nocolor = {"none"}


class SvgGradientCoord:
    def __init__(self, name, comp, value, percent):
        self.name = name
        self.comp = comp
        self.value = value
        self.percent = percent

    def to_value(self, bbox, default=None):
        if self.value is None:
            return default

        if not self.percent:
            return self.value

        if self.comp == "w":
            return (bbox.x2 - bbox.x1) * self.value

        if self.comp == "x":
            return bbox.x1 + (bbox.x2 - bbox.x1) * self.value

        return bbox.y1 + (bbox.y2 - bbox.y1) * self.value

    def parse(self, attr, default_percent):
        if attr is None:
            return
        if attr.endswith("%"):
            self.percent = True
            self.value = float(attr[:-1])/100
        else:
            self.percent = default_percent
            self.value = float(attr)


class SvgGradient:
    def __init__(self):
        self.colors = []
        self.coords = []
        self.matrix = TransformMatrix()

    def add_color(self, offset, color):
        self.colors.append((offset, color[:4]))

    def to_lottie(self, gradient_shape, shape, time=0):
        """!
        @param gradient_shape   Should be a GradientFill or GradientStroke
        @param shape            ShapeElement to apply the gradient to
        @param time             Time to fetch properties from @p shape

        """
        for off, col in self.colors:
            gradient_shape.colors.add_color(off, col)

    def add_coord(self, value):
        setattr(self, value.name, value)
        self.coords.append(value)

    def parse_attrs(self, attrib):
        relunits = attrib.get("gradientUnits", "") != "userSpaceOnUse"
        for c in self.coords:
            c.parse(attrib.get(c.name, None), relunits)


class SvgLinearGradient(SvgGradient):
    def __init__(self):
        super().__init__()
        self.add_coord(SvgGradientCoord("x1", "x", 0, True))
        self.add_coord(SvgGradientCoord("y1", "y", 0, True))
        self.add_coord(SvgGradientCoord("x2", "x", 1, True))
        self.add_coord(SvgGradientCoord("y2", "y", 0, True))

    def to_lottie(self, gradient_shape, shape, time=0):
        bbox = shape.bounding_box(time)
        gradient_shape.start_point.value = self.matrix.apply(NVector(
            self.x1.to_value(bbox),
            self.y1.to_value(bbox),
        ))
        gradient_shape.end_point.value = self.matrix.apply(NVector(
            self.x2.to_value(bbox),
            self.y2.to_value(bbox),
        ))
        gradient_shape.gradient_type = objects.GradientType.Linear

        super().to_lottie(gradient_shape, shape, time)


class SvgRadialGradient(SvgGradient):
    def __init__(self):
        super().__init__()
        self.add_coord(SvgGradientCoord("cx", "x", 0.5, True))
        self.add_coord(SvgGradientCoord("cy", "y", 0.5, True))
        self.add_coord(SvgGradientCoord("fx", "x", None, True))
        self.add_coord(SvgGradientCoord("fy", "y", None, True))
        self.add_coord(SvgGradientCoord("r", "w", 0.5, True))

    def to_lottie(self, gradient_shape, shape, time=0):
        bbox = shape.bounding_box(time)
        cx = self.cx.to_value(bbox)
        cy = self.cy.to_value(bbox)
        gradient_shape.start_point.value = self.matrix.apply(NVector(cx, cy))
        r = self.r.to_value(bbox)
        gradient_shape.end_point.value = self.matrix.apply(NVector(cx+r, cy))

        fx = self.fx.to_value(bbox, cx) - cx
        fy = self.fy.to_value(bbox, cy) - cy
        gradient_shape.highlight_angle.value = math.atan2(fy, fx) * 180 / math.pi
        gradient_shape.highlight_length.value = math.hypot(fx, fy)

        gradient_shape.gradient_type = objects.GradientType.Radial

        super().to_lottie(gradient_shape, shape, time)


def parse_color(color, current_color=Color(0, 0, 0, 1)):
    """!
    Parses CSS colors

    @see https://www.w3.org/wiki/CSS/Properties/color
    """
    # #112233
    if re.match(r"^#[0-9a-fA-F]{6}$", color):
        return Color(int(color[1:3], 16) / 0xff, int(color[3:5], 16) / 0xff, int(color[5:7], 16) / 0xff, 1)
    # #fff
    if re.match(r"^#[0-9a-fA-F]{3}$", color):
        return Color(int(color[1], 16) / 0xf, int(color[2], 16) / 0xf, int(color[3], 16) / 0xf, 1)
    # rgba(123, 123, 123, 0.7)
    match = re.match(r"^rgba\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9.eE]+)\s*\)$", color)
    if match:
        return Color(int(match[1])/255, int(match[2])/255, int(match[3])/255, float(match[4]))
    # rgb(123, 123, 123)
    match = re.match(r"^rgb\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$", color)
    if match:
        return Color(int(match[1])/255, int(match[2])/255, int(match[3])/255, 1)
    # rgb(60%, 30%, 20%)
    match = re.match(r"^rgb\s*\(\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*\)$", color)
    if match:
        return Color(float(match[1])/100, float(match[2])/100, float(match[3])/100, 1)
    # rgba(60%, 30%, 20%, 0.7)
    match = re.match(r"^rgba\s*\(\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)\s*\)$", color)
    if match:
        return Color(float(match[1])/100, float(match[2])/100, float(match[3])/100, float(match[4]))
    # transparent
    if color == "transparent":
        return Color(0, 0, 0, 0)
    # hsl(60, 30%, 20%)
    match = re.match(r"^hsl\s*\(\s*([0-9.eE]+)\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*\)$", color)
    if match:
        return Color(*(colorsys.hls_to_rgb(float(match[1])/360, float(match[3])/100, float(match[2])/100) + (1,)))
    # hsla(60, 30%, 20%, 0.7)
    match = re.match(r"^hsla\s*\(\s*([0-9.eE]+)\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)%\s*,\s*([0-9.eE]+)\s*\)$", color)
    if match:
        return Color(*(colorsys.hls_to_rgb(float(match[1])/360, float(match[3])/100, float(match[2])/100) + (float(match[4]),)))
    # currentColor
    if color in {"currentColor", "inherit"}:
        return current_color.clone()
    # red
    return Color(*color_table[color])


class SvgDefsParent:
    def __init__(self):
        self.items = {}

    def insert(self, dummy, shape):
        self.items[shape.name] = shape

    def __getitem__(self, key):
        return self.items[key]

    def __setitem__(self, key, value):
        self.items[key] = value

    def __contains__(self, key):
        return key in self.items

    @property
    def shapes(self):
        return self


class SvgParser(SvgHandler):
    def __init__(self, name_mode=NameMode.Inkscape):
        self.init_etree()
        self.name_mode = name_mode
        self.current_color = Color(0, 0, 0, 1)
        self.gradients = {}
        self.max_time = 0
        self.defs = SvgDefsParent()
        self.dpi = 96

    def _get_name(self, element, inkscapequal):
        if self.name_mode == NameMode.Inkscape:
            return element.attrib.get(inkscapequal, element.attrib.get("id"))
        return self._get_id(element)

    def _get_id(self, element):
        if self.name_mode != NameMode.NoName:
            return element.attrib.get("id")
        return None

    def _parse_viewbox(self, attrib):
        return map(float, attrib.replace(",", " ").split())

    def parse_etree(self, etree, layer_frames=0, *args, **kwargs):
        animation = objects.Animation(*args, **kwargs)
        self.animation = animation
        self.max_time = 0
        self.document = etree

        svg = etree.getroot()

        self._get_dpi(svg)

        if "width" in svg.attrib and "height" in svg.attrib:
            animation.width = int(round(self._parse_unit(svg.attrib["width"])))
            animation.height = int(round(self._parse_unit(svg.attrib["height"])))
        else:
            _, _, animation.width, animation.height = self._parse_viewbox(svg.attrib["viewBox"])
        animation.name = self._get_name(svg, self.qualified("sodipodi", "docname"))

        if layer_frames:
            for frame in svg:
                if self.unqualified(frame.tag) == "g":
                    layer = objects.ShapeLayer()
                    layer.in_point = self.max_time
                    animation.add_layer(layer)
                    self._parseshape_g(frame, layer, {})
                    self.max_time += layer_frames
                    layer.out_point = self.max_time
        else:
            self._svg_to_layer(animation, svg)

        if self.max_time:
            animation.out_point = self.max_time

        self._fix_viewbox(svg, (layer for layer in animation.layers if not layer.parent_index))

        return animation

    def etree_to_layer(self, animation, etree):
        svg = etree.getroot()
        self._get_dpi(svg)
        layer = self._svg_to_layer(animation, svg)
        self._fix_viewbox(svg, [layer])
        return layer

    def _get_dpi(self, svg):
        self.dpi = float(svg.attrib.get(self.qualified("inkscape", "export-xdpi"), self.dpi))

    def _svg_to_layer(self, animation, svg):
        self.animation = animation
        layer = objects.ShapeLayer()
        animation.add_layer(layer)
        self.parse_children(svg, layer, self.parse_style(svg, {}))
        if self.max_time:
            for sublayer in layer.find_all(objects.Layer):
                sublayer.out_point = self.max_time
        return layer

    def _fix_viewbox(self, svg, layers):
        if "viewBox" in svg.attrib:
            vbx, vby, vbw, vbh = self._parse_viewbox(svg.attrib["viewBox"])
            if vbx != 0 or vby != 0 or vbw != self.animation.width or vbh != self.animation.height:
                for layer in layers:
                    layer.transform.position.value = -NVector(vbx, vby)
                    layer.transform.scale.value = NVector(self.animation.width / vbw, self.animation.height / vbh) * 100

    def _parse_unit(self, value):
        if not isinstance(value, str):
            return value

        mult = 1
        cmin = 2.54
        if value.endswith("px"):
            value = value[:-2]
        elif value.endswith("vw"):
            value = value[:-2]
            mult = self.animation.width * 0.01
        elif value.endswith("vh"):
            value = value[:-2]
            mult = self.animation.height * 0.01
        elif value.endswith("vmin"):
            value = value[:-4]
            mult = min(self.animation.width, self.animation.height) * 0.01
        elif value.endswith("vmax"):
            value = value[:-4]
            mult = max(self.animation.width, self.animation.height) * 0.01
        elif value.endswith("in"):
            value = value[:-2]
            mult = self.dpi
        elif value.endswith("pc"):
            value = value[:-2]
            mult = self.dpi / 6
        elif value.endswith("pt"):
            value = value[:-2]
            mult = self.dpi / 72
        elif value.endswith("cm"):
            value = value[:-2]
            mult = self.dpi / cmin
        elif value.endswith("mm"):
            value = value[:-2]
            mult = self.dpi / cmin / 10
        elif value.endswith("Q"):
            value = value[:-1]
            mult = self.dpi / cmin / 40

        return float(value) * mult

    def parse_color(self, color):
        return parse_color(color, self.current_color)

    def parse_transform(self, element, group, dest_trans):
        bb = group.bounding_box()
        if not bb.isnull():
            itcx = self.qualified("inkscape", "transform-center-x")
            if itcx in element.attrib:
                cx = float(element.attrib[itcx])
                cy = float(element.attrib[self.qualified("inkscape", "transform-center-y")])
                bbx, bby = bb.center()
                cx += bbx
                cy = bby - cy
                dest_trans.anchor_point.value = NVector(cx, cy)
                dest_trans.position.value = NVector(cx, cy)
            #else:
                #c = bb.center()
                #dest_trans.anchor_point.value = c
                #dest_trans.position.value = c.clone()

        if "transform" not in element.attrib:
            return

        matrix = TransformMatrix()
        read_matrix = False

        for t in re.finditer(r"([a-zA-Z]+)\s*\(([^\)]*)\)", element.attrib["transform"]):
            name = t[1]
            params = list(map(float, t[2].strip().replace(",", " ").split()))
            if name == "translate":
                dest_trans.position.value += NVector(
                    params[0],
                    (params[1] if len(params) > 1 else 0),
                )
            elif name == "scale":
                xfac = params[0]
                dest_trans.scale.value[0] = (dest_trans.scale.value[0] / 100 * xfac) * 100
                yfac = params[1] if len(params) > 1 else xfac
                dest_trans.scale.value[1] = (dest_trans.scale.value[1] / 100 * yfac) * 100
            elif name == "rotate":
                ang = params[0]
                x = y = 0
                if len(params) > 2:
                    x = params[1]
                    y = params[2]
                    ap = NVector(x, y)
                    dap = ap - dest_trans.position.value
                    dest_trans.position.value += dap
                    dest_trans.anchor_point.value += dap
                dest_trans.rotation.value = ang
            else:
                read_matrix = True
                self._apply_transform_element_to_matrix(matrix, t)

        if read_matrix:
            dest_trans.position.value -= dest_trans.anchor_point.value
            dest_trans.anchor_point.value = NVector(0, 0)
            trans = matrix.extract_transform()
            dest_trans.skew_axis.value = math.degrees(trans["skew_axis"])
            dest_trans.skew.value = -math.degrees(trans["skew_angle"])
            dest_trans.position.value += trans["translation"]
            dest_trans.rotation.value -= math.degrees(trans["angle"])
            dest_trans.scale.value *= trans["scale"]

    def parse_style(self, element, parent_style):
        style = parent_style.copy()
        for att in css_atrrs & set(element.attrib.keys()):
            if att in element.attrib:
                style[att] = element.attrib[att]
        if "style" in element.attrib:
            style.update(**dict(map(
                lambda x: map(lambda y: y.strip(), x.split(":")),
                filter(bool, element.attrib["style"].split(";"))
            )))
        return style

    def apply_common_style(self, style, transform):
        opacity = float(style.get("opacity", 1))
        transform.opacity.value = opacity * 100

    def apply_visibility(self, style, object):
        if style.get("display", "inline") == "none" or style.get("visibility", "visible") == "hidden":
            object.hidden = True

    def add_shapes(self, element, shapes, shape_parent, parent_style):
        style = self.parse_style(element, parent_style)

        group = objects.Group()
        self.apply_common_style(style, group.transform)
        self.apply_visibility(style, group)
        group.name = self._get_name(element, self.qualified("inkscape", "label"))

        shape_parent.shapes.insert(0, group)
        for shape in shapes:
            group.add_shape(shape)

        self._add_style_shapes(style, group)

        self.parse_transform(element, group, group.transform)

        return group

    def _add_style_shapes(self, style, group):
        stroke_color = style.get("stroke", "none")
        if stroke_color not in nocolor:
            if stroke_color.startswith("url"):
                stroke = self.get_color_url(stroke_color, objects.GradientStroke, group)
                opacity = 1
            else:
                stroke = objects.Stroke()
                color = self.parse_color(stroke_color)
                stroke.color.value = color
                opacity = color[3]
            group.add_shape(stroke)

            stroke.opacity.value = opacity * float(style.get("stroke-opacity", 1)) * 100

            stroke.width.value = self._parse_unit(style.get("stroke-width", 1))

            linecap = style.get("stroke-linecap")
            if linecap == "round":
                stroke.line_cap = objects.shapes.LineCap.Round
            elif linecap == "butt":
                stroke.line_cap = objects.shapes.LineCap.Butt
            elif linecap == "square":
                stroke.line_cap = objects.shapes.LineCap.Square

            linejoin = style.get("stroke-linejoin")
            if linejoin == "round":
                stroke.line_join = objects.shapes.LineJoin.Round
            elif linejoin == "bevel":
                stroke.line_join = objects.shapes.LineJoin.Bevel
            elif linejoin in {"miter", "arcs", "miter-clip"}:
                stroke.line_join = objects.shapes.LineJoin.Miter

            stroke.miter_limit = self._parse_unit(style.get("stroke-miterlimit", 0))

            dash_array = style.get("stroke-dasharray")
            if dash_array and dash_array != "none":
                values = list(map(self._parse_unit, dash_array.replace(",", " ").split()))
                if len(values) % 2:
                    values += values

                stroke.dashes = []
                for i in range(0, len(values), 2):
                    stroke.dashes.append(objects.StrokeDash(values[i], objects.StrokeDashType.Dash))
                    stroke.dashes.append(objects.StrokeDash(values[i+1], objects.StrokeDashType.Gap))

        fill_color = style.get("fill", "inherit")
        if fill_color not in nocolor:
            if fill_color.startswith("url"):
                fill = self.get_color_url(fill_color, objects.GradientFill, group)
                opacity = 1
            else:
                color = self.parse_color(fill_color)
                fill = objects.Fill(color)
                opacity = color[3]
            opacity *= float(style.get("fill-opacity", 1))
            fill.opacity.value = opacity * 100

            if style.get("fill-rule", "") == "evenodd":
                fill.fill_rule = objects.FillRule.EvenOdd

            group.add_shape(fill)

    def _parseshape_use(self, element, shape_parent, parent_style):
        link = element.attrib[self.qualified("xlink", "href")]
        if link.startswith("#"):
            id = link[1:]
            base_element = self.document.find(".//*[@id='%s']" % id)
            use_style = self.parse_style(element, parent_style)
            used = objects.Group()
            shape_parent.add_shape(used)
            used.name = "use"
            used.transform.position.value.x = float(element.attrib.get("x", 0))
            used.transform.position.value.y = float(element.attrib.get("y", 0))
            self.parse_transform(element, used, used.transform)
            self.parse_shape(base_element, used, use_style)
            return used

    def _parseshape_g(self, element, shape_parent, parent_style):
        group = objects.Group()
        shape_parent.shapes.insert(0, group)
        style = self.parse_style(element, parent_style)
        self.apply_common_style(style, group.transform)
        self.apply_visibility(style, group)
        group.name = self._get_name(element, self.qualified("inkscape", "label"))
        self.parse_children(element, group, style)
        self.parse_transform(element, group, group.transform)
        if group.hidden: # Lottie web doesn't seem to support .hd
            group.transform.opacity.value = 0
        return group

    def _parseshape_ellipse(self, element, shape_parent, parent_style):
        ellipse = objects.Ellipse()
        ellipse.position.value = NVector(
            self._parse_unit(element.attrib["cx"]),
            self._parse_unit(element.attrib["cy"])
        )
        ellipse.size.value = NVector(
            self._parse_unit(element.attrib["rx"]) * 2,
            self._parse_unit(element.attrib["ry"]) * 2
        )
        self.add_shapes(element, [ellipse], shape_parent, parent_style)
        return ellipse

    def _parseshape_anim_ellipse(self, ellipse, element, animations):
        self._merge_animations(element, animations, "cx", "cy", "position")
        self._merge_animations(element, animations, "rx", "ry", "size", lambda x, y: NVector(x, y) * 2)
        self._apply_animations(ellipse.position, "position", animations)
        self._apply_animations(ellipse.size, "size", animations)

    def _parseshape_circle(self, element, shape_parent, parent_style):
        ellipse = objects.Ellipse()
        ellipse.position.value = NVector(
            self._parse_unit(element.attrib["cx"]),
            self._parse_unit(element.attrib["cy"])
        )
        r = self._parse_unit(element.attrib["r"]) * 2
        ellipse.size.value = NVector(r, r)
        self.add_shapes(element, [ellipse], shape_parent, parent_style)
        return ellipse

    def _parseshape_anim_circle(self, ellipse, element, animations):
        self._merge_animations(element, animations, "cx", "cy", "position")
        self._apply_animations(ellipse.position, "position", animations)
        self._apply_animations(ellipse.size, "r", animations, lambda r: NVector(r, r) * 2)

    def _parseshape_rect(self, element, shape_parent, parent_style):
        rect = objects.Rect()
        w = self._parse_unit(element.attrib.get("width", 0))
        h = self._parse_unit(element.attrib.get("height", 0))
        rect.position.value = NVector(
            self._parse_unit(element.attrib.get("x", 0)) + w / 2,
            self._parse_unit(element.attrib.get("y", 0)) + h / 2
        )
        rect.size.value = NVector(w, h)
        rx = self._parse_unit(element.attrib.get("rx", 0))
        ry = self._parse_unit(element.attrib.get("ry", 0))
        rect.rounded.value = (rx + ry) / 2
        self.add_shapes(element, [rect], shape_parent, parent_style)
        return rect

    def _parseshape_anim_rect(self, rect, element, animations):
        self._merge_animations(element, animations, "width", "height", "size", lambda x, y: NVector(x, y))
        self._apply_animations(rect.size, "size", animations)
        self._merge_animations(element, animations, "x", "y", "position")
        self._merge_animations(element, animations, "position", "size", "position", lambda p, s: p + s / 2)
        self._apply_animations(rect.position, "position", animations)
        self._merge_animations(element, animations, "rx", "ry", "rounded", lambda x, y: (x + y) / 2)
        self._apply_animations(rect.rounded, "rounded", animations)

    def _parseshape_line(self, element, shape_parent, parent_style):
        line = objects.Path()
        line.shape.value.add_point(NVector(
            self._parse_unit(element.attrib["x1"]),
            self._parse_unit(element.attrib["y1"])
        ))
        line.shape.value.add_point(NVector(
            self._parse_unit(element.attrib["x2"]),
            self._parse_unit(element.attrib["y2"])
        ))
        return self.add_shapes(element, [line], shape_parent, parent_style)

    def _parseshape_anim_line(self, group, element, animations):
        line = group.shapes[0]

        self._merge_animations(element, animations, "x1", "y1", "p1")
        self._merge_animations(element, animations, "x2", "y2", "p2")

        def combine(p1, p2):
            shape = objects.Bezier()
            shape.add_point(p1)
            shape.add_point(p2)
            return shape

        self._merge_animations(element, animations, "p1", "p2", "points", combine)
        self._apply_animations(line.shape, "points", animations)

    def _handle_poly(self, element):
        line = objects.Path()
        coords = list(map(float, element.attrib["points"].replace(",", " ").split()))
        for i in range(0, len(coords), 2):
            line.shape.value.add_point(NVector(*coords[i:i+2]))
        return line

    def _parseshape_polyline(self, element, shape_parent, parent_style):
        line = self._handle_poly(element)
        return self.add_shapes(element, [line], shape_parent, parent_style)

    def _parseshape_polygon(self, element, shape_parent, parent_style):
        line = self._handle_poly(element)
        line.shape.value.close()
        return self.add_shapes(element, [line], shape_parent, parent_style)

    def _parseshape_path(self, element, shape_parent, parent_style):
        d_parser = PathDParser(element.attrib.get("d", ""))
        d_parser.parse()
        paths = []
        for path in d_parser.paths:
            p = objects.Path()
            p.shape.value = path
            paths.append(p)
        #if len(d_parser.paths) > 1:
            #paths.append(objects.shapes.Merge())
        return self.add_shapes(element, paths, shape_parent, parent_style)

    def parse_children(self, element, shape_parent, parent_style):
        for child in element:
            tag = self.unqualified(child.tag)
            if not self.parse_shape(child, shape_parent, parent_style):
                handler = getattr(self, "_parse_" + tag, None)
                if handler:
                    handler(child)

    def parse_shape(self, element, shape_parent, parent_style):
        handler = getattr(self, "_parseshape_" + self.unqualified(element.tag), None)
        if handler:
            out = handler(element, shape_parent, parent_style)
            self.parse_animations(out, element)
            if element.attrib.get("id"):
                self.defs.items[element.attrib["id"]] = out
            return out
        return None

    def _parse_defs(self, element):
        self.parse_children(element, self.defs, {})

    def _apply_transform_element_to_matrix(self, matrix, t):
        name = t[1]
        params = list(map(float, t[2].strip().replace(",", " ").split()))
        if name == "translate":
            matrix.translate(
                params[0],
                (params[1] if len(params) > 1 else 0),
            )
        elif name == "scale":
            xfac = params[0]
            yfac = params[1] if len(params) > 1 else xfac
            matrix.scale(xfac, yfac)
        elif name == "rotate":
            ang = params[0]
            x = y = 0
            if len(params) > 2:
                x = params[1]
                y = params[2]
                matrix.translate(-x, -y)
                matrix.rotate(math.radians(ang))
                matrix.translate(x, y)
            else:
                matrix.rotate(math.radians(ang))
        elif name == "skewX":
            matrix.skew(math.radians(params[0]), 0)
        elif name == "skewY":
            matrix.skew(0, math.radians(params[0]))
        elif name == "matrix":
            m = TransformMatrix()
            m.a, m.b, m.c, m.d, m.tx, m.ty = params
            matrix *= m

    def _transform_to_matrix(self, transform):
        matrix = TransformMatrix()

        for t in re.finditer(r"([a-zA-Z]+)\s*\(([^\)]*)\)", transform):
            self._apply_transform_element_to_matrix(matrix, t)

        return matrix

    def _gradient(self, element, grad):
        grad.matrix = self._transform_to_matrix(element.attrib.get("gradientTransform", ""))

        id = element.attrib["id"]
        if id in self.gradients:
            grad.colors = self.gradients[id].colors
        grad.parse_attrs(element.attrib)
        href = element.attrib.get(self.qualified("xlink", "href"))
        if href:
            srcid = href.strip("#")
            if srcid in self.gradients:
                src = self.gradients[srcid]
            else:
                src = grad.__class__()
                self.gradients[srcid] = src
            grad.colors = src.colors

        for stop in element.findall("./%s" % self.qualified("svg", "stop")):
            offset = stop.attrib.get("offset", "0")
            off = float(offset.strip("%"))
            if offset.endswith("%"):
                off /= 100
            style = self.parse_style(stop, {})
            color = self.parse_color(style.get("stop-color", "black"))
            if "stop-opacity" in style:
                color[3] = float(style["stop-opacity"])
            grad.add_color(off, color)
        self.gradients[id] = grad

    def _parse_linearGradient(self, element):
        self._gradient(element, SvgLinearGradient())

    def _parse_radialGradient(self, element):
        self._gradient(element, SvgRadialGradient())

    def get_color_url(self, color, gradientclass, shape):
        match = re.match(r"""url\(['"]?#([^)'"]+)['"]?\)""", color)
        if not match:
            return None
        id = match[1]
        if id not in self.gradients:
            return None
        grad = self.gradients[id]
        outgrad = gradientclass()
        grad.to_lottie(outgrad, shape)
        if self.name_mode != NameMode.NoName:
            grad.name = id
        return outgrad

    ## @todo Parse single font property, fallback family etc
    def _parse_text_style(self, style, font_style=None):
        if "font-family" in style:
            font_style.query.family(style["font-family"].strip("'\""))

        if "font-style" in style:
            if style["font-style"] == "oblique":
                font_style.query.custom("slant", 110)
            elif style["font-style"] == "italic":
                font_style.query.custom("slant", 100)

        if "font-weight" in style:
            if style["font-weight"] in {"bold", "bolder"}:
                font_style.query.weight(200)
            elif style["font-weight"] == "lighter":
                font_style.query.weight(50)
            elif style["font-weight"].isdigit():
                font_style.query.css_weight(int(style["font-weight"]))

        if "font-size" in style:
            fz = style["font-size"]
            fz_names = {
                "xx-small": 8,
                "x-small": 16,
                "small": 32,
                "medium": 64,
                "large": 128,
                "x-large": 256,
                "xx-large": 512,
            }
            if fz in fz_names:
                font_style.size = fz_names[fz]
            elif fz == "smaller":
                font_style.size /= 2
            elif fz == "larger":
                font_style.size *= 2
            elif fz.endswith("px"):
                font_style.size = float(fz[:-2])
            elif fz.isnumeric():
                font_style.size = float(fz)

        if "text-align" in style:
            ta = style["text-align"]
            if ta in ("left", "start"):
                font_style.justify = font.TextJustify.Left
            elif ta == "center":
                font_style.justify = font.TextJustify.Center
            elif ta in ("right", "end"):
                font_style.justify = font.TextJustify.Right

    def _parse_text_elem(self, element, style, group, parent_style, font_style):
        self._parse_text_style(style, font_style)

        if "x" in element.attrib or "y" in element.attrib:
            font_style.position = NVector(
                float(element.attrib["x"]),
                float(element.attrib["y"]),
            )

        childpos = NVector(0, font_style.position.y)

        if element.text:
            fs = font.FontShape(element.text, font_style)
            fs.refresh()
            group.add_shape(fs)
            childpos.x = fs.wrapped.next_x

        for child in element:
            if child.tag == self.qualified("svg", "tspan"):
                child_style = font_style.clone()
                child_style.position = childpos.clone()
                fs = self._parseshape_text(child, group, parent_style, child_style)
                childpos.x = fs.next_x
            if child.tail:
                child_style = font_style.clone()
                child_style.position = childpos.clone()
                fs = font.FontShape(child.tail, child_style)
                fs.refresh()
                group.add_shape(fs)
                childpos.x = fs.wrapped.next_x

        group.next_x = childpos.x

    def _parseshape_text(self, element, shape_parent, parent_style, font_style=None):
        group = objects.Group()

        style = self.parse_style(element, parent_style)
        self.apply_common_style(style, group.transform)
        self.apply_visibility(style, group)
        group.name = self._get_id(element)

        if has_font:
            if font_style is None:
                font_style = font.FontStyle("", 64)
            self._parse_text_elem(element, style, group, style, font_style)

        style.setdefault("fill", "none")
        self._add_style_shapes(style, group)

        ## @todo text-anchor when it doesn't match text-align
        #if element.tag == self.qualified("svg", "text"):
            #dx = 0
            #dy = 0

            #ta = style.get("text-anchor", style.get("text-align", ""))
            #if ta == "middle":
                #dx -= group.bounding_box().width / 2
            #elif ta == "end":
                #dx -= group.bounding_box().width

            #if dx or dy:
                #ng = objects.Group()
                #ng.add_shape(group)
                #group.transform.position.value.x += dx
                #group.transform.position.value.y += dy
                #group = ng

        shape_parent.shapes.insert(0, group)
        self.parse_transform(element, group, group.transform)
        return group

    def parse_animations(self, lottie, element):
        animations = {}
        for child in element:
            if self.unqualified(child.tag) == "animate":
                att = child.attrib["attributeName"]

                from_val = child.attrib["from"]
                if att == "d":
                    ## @todo
                    continue
                else:
                    from_val = float(from_val)
                    if "to" in child.attrib:
                        to_val = float(child.attrib["to"])
                    elif "by" in child.attrib:
                        to_val = float(child.attrib["by"]) + from_val

                begin = self.parse_animation_time(child.attrib.get("begin", 0)) or 0
                if "dur" in child.attrib:
                    end = (self.parse_animation_time(child.attrib["dur"]) or 0) + begin
                elif "end" in child.attrib:
                    end = self.parse_animation_time(child.attrib["dur"]) or 0
                else:
                    continue

                if att not in animations:
                    animations[att] = {}
                animations[att][begin] = from_val
                animations[att][end] = to_val
                if self.max_time < end:
                    self.max_time = end

        tag = self.unqualified(element.tag)
        handler = getattr(self, "_parseshape_anim_" + tag, None)
        if handler:
            handler(lottie, element, animations)

    def parse_animation_time(self, value):
        """!
        @see https://developer.mozilla.org/en-US/docs/Web/SVG/Content_type#Clock-value
        """
        if not value:
            return None
        try:
            seconds = 0
            if ":" in value:
                mult = 1
                for elem in reversed(value.split(":")):
                    seconds += float(elem) * mult
                    mult *= 60
            elif value.endswith("s"):
                seconds = float(value[:-1])
            elif value.endswith("ms"):
                seconds = float(value[:-2]) / 1000
            elif value.endswith("min"):
                seconds = float(value[:-3]) * 60
            elif value.endswith("h"):
                seconds = float(value[:-1]) * 60 * 60
            else:
                seconds = float(value)
            return seconds * self.animation.frame_rate
        except ValueError:
            pass
        return None

    def _merge_animations(self, element, animations, val1, val2, dest, merge=NVector):
        if val1 not in animations and val2 not in animations:
            return

        dict1 = list(sorted(animations.pop(val1, {}).items()))
        dict2 = list(sorted(animations.pop(val2, {}).items()))

        x = float(element.attrib[val1])
        y = float(element.attrib[val2])
        values = {}
        while dict1 or dict2:
            if not dict1 or (dict2 and dict1[0][0] > dict2[0][0]):
                t, y = dict2.pop(0)
            elif not dict2 or dict1[0][0] < dict2[0][0]:
                t, x = dict1.pop(0)
            else:
                t, x = dict1.pop(0)
                t, y = dict2.pop(0)

            values[t] = merge(x, y)

        animations[dest] = values

    def _apply_animations(self, animatable, name, animations, transform=lambda v: v):
        if name in animations:
            for t, v in animations[name].items():
                animatable.add_keyframe(t, transform(v))


class PathDParser:
    _re = re.compile("|".join((
        r"[a-zA-Z]",
        r"[-+]?[0-9]*\.?[0-9]*[eE][-+]?[0-9]+",
        r"[-+]?[0-9]*\.?[0-9]+",
    )))

    def __init__(self, d_string):
        self.path = objects.properties.Bezier()
        self.paths = []
        self.p = NVector(0, 0)
        self.la = None
        self.la_type = None
        self.tokens = list(map(self.d_subsplit, self._re.findall(d_string)))
        self.add_p = True
        self.implicit = "M"

    def d_subsplit(self, tok):
        if tok.isalpha():
            return tok
        return float(tok)

    def next_token(self):
        if self.tokens:
            self.la = self.tokens.pop(0)
            if isinstance(self.la, str):
                self.la_type = 0
            else:
                self.la_type = 1
        else:
            self.la = None
        return self.la

    def next_vec(self):
        x = self.next_token()
        y = self.next_token()
        return NVector(x, y)

    def cur_vec(self):
        x = self.la
        y = self.next_token()
        return NVector(x, y)

    def parse(self):
        self.next_token()
        while self.la is not None:
            if self.la_type == 0:
                parser = "_parse_" + self.la
                self.next_token()
                getattr(self, parser)()
            else:
                parser = "_parse_" + self.implicit
                getattr(self, parser)()

    def _push_path(self):
        self.path = objects.properties.Bezier()
        self.add_p = True

    def _parse_M(self):
        if self.la_type != 1:
            self.next_token()
            return
        self.p = self.cur_vec()
        self.implicit = "L"
        if not self.add_p:
            self._push_path()
        self.next_token()

    def _parse_m(self):
        if self.la_type != 1:
            self.next_token()
            return
        self.p += self.cur_vec()
        self.implicit = "l"
        if not self.add_p:
            self._push_path()
        self.next_token()

    def _rpoint(self, point, rel=None):
        return (point - (rel or self.p)) if point is not None else NVector(0, 0)

    def _do_add_p(self, outp=None):
        if self.add_p:
            self.paths.append(self.path)
            self.path.add_point(self.p.clone(), NVector(0, 0), self._rpoint(outp))
            self.add_p = False
        elif outp:
            rp = self.path.vertices[-1]
            self.path.out_tangents[-1] = self._rpoint(outp, rp)

    def _parse_L(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p = self.cur_vec()
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "L"
        self.next_token()

    def _parse_l(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p += self.cur_vec()
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "l"
        self.next_token()

    def _parse_H(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[0] = self.la
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "H"
        self.next_token()

    def _parse_h(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[0] += self.la
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "h"
        self.next_token()

    def _parse_V(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[1] = self.la
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "V"
        self.next_token()

    def _parse_v(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[1] += self.la
        self.path.add_point(self.p.clone(), NVector(0, 0), NVector(0, 0))
        self.implicit = "v"
        self.next_token()

    def _parse_C(self):
        if self.la_type != 1:
            self.next_token()
            return
        pout = self.cur_vec()
        self._do_add_p(pout)
        pin = self.next_vec()
        self.p = self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "C"
        self.next_token()

    def _parse_c(self):
        if self.la_type != 1:
            self.next_token()
            return
        pout = self.p + self.cur_vec()
        self._do_add_p(pout)
        pin = self.p + self.next_vec()
        self.p += self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "c"
        self.next_token()

    def _parse_S(self):
        if self.la_type != 1:
            self.next_token()
            return
        pin = self.cur_vec()
        self._do_add_p()
        handle = self.path.in_tangents[-1]
        self.path.out_tangents[-1] = (-handle)
        self.p = self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "S"
        self.next_token()

    def _parse_s(self):
        if self.la_type != 1:
            self.next_token()
            return
        pin = self.cur_vec() + self.p
        self._do_add_p()
        handle = self.path.in_tangents[-1]
        self.path.out_tangents[-1] = (-handle)
        self.p += self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "s"
        self.next_token()

    def _parse_Q(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        pin = self.cur_vec()
        self.p = self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "Q"
        self.next_token()

    def _parse_q(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        pin = self.p + self.cur_vec()
        self.p += self.next_vec()
        self.path.add_point(
            self.p.clone(),
            (pin - self.p),
            NVector(0, 0)
        )
        self.implicit = "q"
        self.next_token()

    def _parse_T(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        handle = self.p - self.path.in_tangents[-1]
        self.p = self.cur_vec()
        self.path.add_point(
            self.p.clone(),
            (handle - self.p),
            NVector(0, 0)
        )
        self.implicit = "T"
        self.next_token()

    def _parse_t(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        handle = -self.path.in_tangents[-1] + self.p
        self.p += self.cur_vec()
        self.path.add_point(
            self.p.clone(),
            (handle - self.p),
            NVector(0, 0)
        )
        self.implicit = "t"
        self.next_token()

    def _parse_A(self):
        if self.la_type != 1:
            self.next_token()
            return
        r = self.cur_vec()
        xrot = self.next_token()
        large = self.next_token()
        sweep = self.next_token()
        dest = self.next_vec()
        self._do_arc(r[0], r[1], xrot, large, sweep, dest)
        self.implicit = "A"
        self.next_token()

    def _do_arc(self, rx, ry, xrot, large, sweep, dest):
        self._do_add_p()
        if self.p == dest:
            return

        if rx == 0 or ry == 0:
            # Straight line
            self.p = dest
            self.path.add_point(
                self.p.clone(),
                NVector(0, 0),
                NVector(0, 0)
            )
            return

        ellipse, theta1, deltatheta = Ellipse.from_svg_arc(self.p, rx, ry, xrot, large, sweep, dest)
        points = ellipse.to_bezier_points(theta1, deltatheta)

        self._do_add_p()
        self.path.out_tangents[-1] = points[0].out_tangent
        for point in points[1:-1]:
            self.path.add_point(
                point.vertex,
                point.in_tangent,
                point.out_tangent,
            )
        self.path.add_point(
            dest.clone(),
            points[-1].in_tangent,
            NVector(0, 0),
        )
        self.p = dest

    def _parse_a(self):
        if self.la_type != 1:
            self.next_token()
            return
        r = self.cur_vec()
        xrot = self.next_token()
        large = self.next_token()
        sweep = self.next_token()
        dest = self.p + self.next_vec()
        self._do_arc(r[0], r[1], xrot, large, sweep, dest)
        self.implicit = "a"
        self.next_token()

    def _parse_Z(self):
        if self.path.vertices:
            self.p = self.path.vertices[0].clone()
        self.path.close()
        self._push_path()

    def _parse_z(self):
        self._parse_Z()


def parse_svg_etree(etree, layer_frames=0, *args, **kwargs):
    parser = SvgParser()
    return parser.parse_etree(etree, layer_frames, *args, **kwargs)


def parse_svg_file(file, layer_frames=0, *args, **kwargs):
    return parse_svg_etree(ElementTree.parse(file), layer_frames, *args, **kwargs)



#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie import __version__

parser = argparse.ArgumentParser(
    description="Shows the palette of a raster image"
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "infile",
    help="Input file"
)
parser.add_argument(
    "--colors", "-c",
    default=1,
    type=int,
    help="Number of colors to quantize"
)

if __name__ == "__main__":
    from lottie.parsers.raster import RasterImage
    ns = parser.parse_args()

    raster = RasterImage.open(ns.infile)
    palette = raster.k_means(ns.colors)

    for color in palette:
        print("#%02x%02x%02x %f : %s" % (
            int(round(color[0])),
            int(round(color[1])),
            int(round(color[2])),
            color[3] / 255,
            list(color/255),
        ))



import math
from ..nvector import NVector


def _sign(x):
    if x < 0:
        return -1
    return 1


class TransformMatrix:
    scalar = float

    def __init__(self):
        """ Creates an Identity matrix """
        self.to_identity()

    def clone(self):
        m = TransformMatrix()
        m._mat = list(self._mat)
        return m

        return self

    def __getitem__(self, key):
        row, col = key
        return self._mat[row*4+col]

    def __setitem__(self, key, value):
        row, col = key
        self._mat[row*4+col] = self.scalar(value)

    @property
    def a(self):
        return self[0, 0]

    @a.setter
    def a(self, v):
        self[0, 0] = self.scalar(v)

    @property
    def b(self):
        return self[0, 1]

    @b.setter
    def b(self, v):
        self[0, 1] = self.scalar(v)

    @property
    def c(self):
        return self[1, 0]

    @c.setter
    def c(self, v):
        self[1, 0] = self.scalar(v)

    @property
    def d(self):
        return self[1, 1]

    @d.setter
    def d(self, v):
        self[1, 1] = self.scalar(v)

    @property
    def tx(self):
        return self[3, 0]

    @tx.setter
    def tx(self, v):
        self[3, 0] = self.scalar(v)

    @property
    def ty(self):
        return self[3, 1]

    @ty.setter
    def ty(self, v):
        self[3, 1] = self.scalar(v)

    def __str__(self):
        return str(self._mat)

    def scale(self, x, y=None):
        if y is None:
            y = x

        m = TransformMatrix()
        m.a = x
        m.d = y
        self *= m
        return self

    def translate(self, x, y=None):
        if y is None:
            x, y = x
        m = TransformMatrix()
        m.tx = x
        m.ty = y
        self *= m
        return self

    def skew(self, x_rad, y_rad):
        m = TransformMatrix()
        m.c = math.tan(x_rad)
        m.b = math.tan(y_rad)
        self *= m
        return self

    def skew_from_axis(self, skew, axis):
        self.rotate(axis)
        m = TransformMatrix()
        m.c = math.tan(skew)
        self *= m
        self.rotate(-axis)
        return self

    def row(self, i):
        return NVector(self[i, 0], self[i, 1], self[i, 2], self[i, 3])

    def column(self, i):
        return NVector(self[0, i], self[1, i], self[2, i], self[3, i])

    def to_identity(self):
        self._mat = [
            1., 0., 0., 0.,
            0., 1., 0., 0.,
            0., 0., 1., 0.,
            0., 0., 0., 1.,
        ]

    def apply(self, vector):
        vector3 = NVector(vector.x, vector.y, 0, 1)
        return NVector(
            self.column(0).dot(vector3),
            self.column(1).dot(vector3),
        )

    @classmethod
    def rotation(cls, radians):
        m = cls()
        m.a = math.cos(radians)
        m.b = -math.sin(radians)
        m.c = math.sin(radians)
        m.d = math.cos(radians)

        return m

    def __mul__(self, other):
        m = TransformMatrix()
        for row in range(4):
            for col in range(4):
                m[row, col] = self.row(row).dot(other.column(col))
        return m

    def __imul__(self, other):
        m = self * other
        self._mat = m._mat
        return self

    def rotate(self, radians):
        self *= TransformMatrix.rotation(radians)
        return self

    def extract_transform(self):
        a = self.a
        b = self.b
        c = self.c
        d = self.d
        tx = self.tx
        ty = self.ty

        dest_trans = {
            "translation": NVector(tx, ty),
            "angle": 0,
            "scale": NVector(1, 1),
            "skew_axis": 0,
            "skew_angle": 0,
        }

        delta = a * d - b * c
        if a != 0 or b != 0:
            r = math.hypot(a, b)
            dest_trans["angle"] = - _sign(b) * math.acos(a/r)
            sx = r
            sy = delta / r
            dest_trans["skew_axis"] = 0
        else:
            r = math.hypot(c, d)
            dest_trans["angle"] = math.pi / 2 + _sign(d) * math.acos(c / r)
            sx = delta / r
            sy = r
            dest_trans["skew_axis"] = math.pi / 2

        dest_trans["scale"] = NVector(sx, sy)

        skew = math.atan2((a * c + b * d), r * r)
        dest_trans["skew_angle"] = skew

        return dest_trans

    def to_css_2d(self):
        return "matrix(%s, %s, %s, %s, %s, %s)" % (
            self.a, self.b, self.c, self.d, self.tx, self.ty
        )
